#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : relu
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-06-10 15:22
   Description : description what the main function of this file
"""

from stensorflow.basic.basic_class.pair import SharedPair
from stensorflow.basic.basic_class.private import PrivateTensor
from stensorflow.basic.operator.selectshare import select_share
from typing import Union
from stensorflow.global_var import StfConfig
from stensorflow.exception.exception import StfTypeException, StfValueException
from stensorflow.basic.protocol.msb import msb, msb_log_round
from stensorflow.basic.operator.order import is_positive
import tensorflow as tf


def drelu_binary_const(x: SharedPair):
    y = is_positive(x)
    return SharedPair.from_SharedPairBase(y)


def drelu_binary_log(x: SharedPair):
    t = msb_log_round(x)
    return t.ones_like() - t


def drelu_binary_linear(x: SharedPair):
    t = msb(x)
    # print("t=", t)
    s = t.ones_like() - t
    # print("s=", s)
    return s


def drelu_binary(x: SharedPair):
    if StfConfig.drelu == "const":
        return drelu_binary_const(x)
    elif StfConfig.drelu == "log":
        return drelu_binary_log(x)
    elif StfConfig.drelu == "linear":
        return drelu_binary_linear(x)
    else:
        raise StfValueException("StfConfig.delay", "high or middle or low", StfConfig.delay)


def relu_local(x: PrivateTensor):
    z = PrivateTensor(owner=x.owner)
    with tf.device(x.owner):
        x = x.to_tf_tensor()
        y = tf.nn.relu(x)
        z.load_from_tf_tensor(y)
    return z


def drelu_local(x: PrivateTensor):
    z = PrivateTensor(owner=x.owner)
    with tf.device(x.owner):
        x = x.to_tf_tensor()
        y = tf.cast(x >= 0, 'float32')
        z.load_from_tf_tensor(y)
    return z


def relu(x: Union[SharedPair, PrivateTensor], drelu_b=None):
    if isinstance(x, SharedPair):
        if drelu_b is not None:
            s = drelu_b
        else:
            s = drelu_binary(x)
        y = select_share(s, x)
        return SharedPair.from_SharedPairBase(y)
    elif isinstance(x, PrivateTensor):
        return relu_local(x)
    else:
        raise StfTypeException("x", "SharedPair or PrivateTensor", type(x))


def relu_pull_back(x: Union[SharedPair, PrivateTensor], ploss_py, drelu_b=None):
    if drelu_b is not None:
        ploss_px = select_share(drelu_b, ploss_py)
        return SharedPair.from_SharedPairBase(ploss_px)
    elif isinstance(x, SharedPair):
        s = drelu_binary(x)
        ploss_px = select_share(s, ploss_py)
        return SharedPair.from_SharedPairBase(ploss_px)
    elif isinstance(x, PrivateTensor):
        return drelu_local(x) * ploss_py
    else:
        raise StfTypeException("x", "SharedPair or PrivateTensor", type(x))




#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : Loss
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-09-11 21:26
   Description : description what the main function of this file
"""

from stensorflow.ml.nn.layers.layer import Layer
from stensorflow.basic.operator.softmax import softmax
from stensorflow.basic.operator.sigmoid import sigmoid_local, sigmoid_sin as sigmoid
from stensorflow.basic.basic_class.pair import SharedPair

class Loss(Layer):
    BinaryCrossEntropyLossWithSigmoid = 0
    CrossEntropyLossWithSoftmax = 1

    def __init__(self, fathers):
        super(Loss, self).__init__(output_dim=[1], fathers=fathers)


class BinaryCrossEntropyLossWithSigmoid(Loss):
    def __init__(self, layer_score, layer_label):

        fathers = [layer_score, layer_label]
        super(BinaryCrossEntropyLossWithSigmoid, self).__init__(fathers=fathers)
        self.M = 512

    def forward(self):
        for father in self.fathers:
            if isinstance(father, Layer):
                father.forward()
            else:
                raise Exception("father must be a layer")
        self.x = list(map(lambda father: father.y, self.fathers))

        self.score = self.x[0]
        self.label = self.x[1]
        self.y = sigmoid(self.score, M=self.M).dup_with_precision(new_fixedpoint=self.score.fixedpoint)

    def backward(self):
        self.ploss_pw = []
        self.ploss_px = {self.fathers[0]: self.y - self.label, self.fathers[1]: -self.score}


class BinaryCrossEntropyLossWithSigmoidLocal(Loss):
    def __init__(self, layer_score: Layer, layer_label: Layer, owner):

        fathers = [layer_score, layer_label]
        super(BinaryCrossEntropyLossWithSigmoidLocal, self).__init__(fathers=fathers)
        self.owner = owner

    def forward(self):

        for father in self.fathers:
            if isinstance(father, Layer):
                father.forward()
            else:
                raise Exception("father must be a layer")
        self.x = list(map(lambda father: father.y, self.fathers))

        self.label = self.x[1]
        self.score = self.x[0].to_private(owner=self.owner)

        self.y = sigmoid_local(self.score).dup_with_precision(new_fixedpoint=self.score.fixedpoint)

    def backward(self):
        self.ploss_pw = []
        self.ploss_px = {self.fathers[0]: self.y - self.label, self.fathers[1]: -self.score}


class CrossEntropyLossWithSoftmax(Loss):
    def __init__(self, layer_score, layer_label):

        fathers = [layer_score, layer_label]
        super(CrossEntropyLossWithSoftmax, self).__init__(fathers=fathers)

    def forward(self):
        for father in self.fathers:
            if isinstance(father, Layer):
                father.forward()
            else:
                raise Exception("father must be a layer")
        self.x = list(map(lambda father: father.y, self.fathers))
        # dense层的输出值
        self.score = self.x[0]
        # 原始label信息
        self.label = self.x[1]
        self.y = softmax(self.score)

    def backward(self):
        self.ploss_pw = []
        self.ploss_px = {self.fathers[0]: self.y - self.label, self.fathers[1]: -self.score}



class CrossEntropyLossWithSoftmax_bak(Loss):
    def __init__(self, layer_score, layer_label):

        fathers = [layer_score, layer_label]
        super(CrossEntropyLossWithSoftmax, self).__init__(fathers=fathers)

    def forward(self):
        for father in self.fathers:
            if isinstance(father, Layer):
                father.forward()
            else:
                raise Exception("father must be a layer")
        self.x = list(map(lambda father: father.y, self.fathers))
        # dense层的输出值
        self.score = self.x[0]
        # 原始label信息
        self.label = self.x[1]
        self.y = softmax(self.score)
        self.ones = self.score.ones_like()


    def backward(self):
        self.ploss_pw = []
        self.ploss_px = {self.fathers[0]: 0.001*(self.ones - 2 * self.label), self.fathers[1]: -self.score}


class CrossEntropyLossWithSoftmaxLocal(Loss):
    def __init__(self, layer_score: Layer, layer_label: Layer, owner):

        fathers = [layer_score, layer_label]
        super(CrossEntropyLossWithSoftmaxLocal, self).__init__(fathers=fathers)
        self.owner = owner

    def forward(self):
        for father in self.fathers:
            if isinstance(father, Layer):
                father.forward()
            else:
                raise Exception("father must be a layer")
        self.x = list(map(lambda father: father.y, self.fathers))

        self.score = self.x[0].to_private(owner=self.owner)
        self.label = self.x[1]
        self.y = softmax(self.score)

    def backward(self):
        self.ploss_pw = []
        self.ploss_px = {self.fathers[0]: self.y - self.label, self.fathers[1]: -self.score}



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group Copyright (c) 2004-2020 All Rights Reserved.
"""

import tensorflow as tf
from tensorflow.python.util.compat import collections_abc
from stensorflow.basic.basic_class.base import SharedTensorBase
from stensorflow.basic.basic_class.private import PrivateTensor, PrivateTensorBase
from stensorflow.basic.basic_class.pair import SharedVariablePair, SharedPair, SharedPairBase
from stensorflow.basic.operator.relu import drelu_binary
from stensorflow.basic.operator.selectshare import select_share
from typing import Union
from stensorflow.global_var import StfConfig
from stensorflow.basic.operator.sigmoid import sin2pi
import numpy as np
PModel = StfConfig.pool_module


def get_sequence(value, n, channel_index, name):
    """
    Formats a value input for avg_pool2d.
    :param value:
    :param n:
    :param channel_index:
    :param name:
    :return:
    """
    if value is None:
        value = [1]
    elif not isinstance(value, collections_abc.Sized):
        value = [value]

    current_n = len(value)
    if current_n == n + 2:
        return value
    elif current_n == 1:
        value = list((value[0],) * n)
    elif current_n == n:
        value = list(value)
    else:
        raise ValueError("{} should be of length 1, {} or {} but was {}".format(
            name, n, n + 2, current_n))

    if channel_index == 1:
        return [1, 1] + value
    else:
        return [1] + value + [1]


def avg_pool2d(input: Union[PrivateTensor, SharedPair],
               ksize,
               strides,
               padding,
               data_format="NHWC"):
    """
    See tf.avg_pool2d
      """
    if data_format is None:
        data_format = "NHWC"
    channel_index = 1 if data_format.startswith("NC") else 3
    ksize = get_sequence(ksize, 2, channel_index, "ksize")
    strides = get_sequence(strides, 2, channel_index, "strides")
    if isinstance(input, PrivateTensorBase):
        with tf.device(input.owner):
            v = PModel.int64_avg_pool(input.inner_value, ksize=ksize, strides=strides, padding=padding)
            z = PrivateTensor(owner=input.owner, inner_value=v, fixedpoint=input.fixedpoint)
        return z
    elif isinstance(input, SharedPairBase):
        with tf.device(input.ownerL):
            vL = PModel.int64_avg_pool(input.xL.inner_value, ksize=ksize, strides=strides, padding=padding)
        with tf.device(input.ownerR):
            vR = PModel.int64_avg_pool(input.xR.inner_value, ksize=ksize, strides=strides, padding=padding)
        z = SharedPair(ownerL=input.ownerL,
                       ownerR=input.ownerR,
                       xL=SharedTensorBase(inner_value=vL, module=input.xL.module, shape=list(vL.shape)),
                       xR=SharedTensorBase(inner_value=vR, module=input.xR.module, shape=list(vR.shape)),
                       fixedpoint=input.fixedpoint)
        return z

    else:
        raise TypeError(
            "Expected PrivateTensorBase or SharedPair  for 'input' argument to 'average' Op, not %r." % input)


def sum_pool2d_grad(input_shape,
                    out_backprop: Union[SharedPair, PrivateTensor],
                    ksize,
                    strides,
                    padding,
                    data_format="NHWC"):
    """
    The operation to compute SumPool gradients.
    :param input_shape: See tf.avg_pool2d_grad
    :param out_backprop: See tf.avg_pool2d_grad
    :param ksize: See tf.avg_pool2d_grad
    :param strides: See tf.avg_pool2d_grad
    :param padding: See tf.avg_pool2d_grad
    :param data_format: See tf.avg_pool2d_grad
    :return:
    """
    if data_format is None:
        data_format = "NHWC"
    channel_index = 1 if data_format.startswith("NC") else 3
    ksize = get_sequence(ksize, 2, channel_index, "ksize")
    strides = get_sequence(strides, 2, channel_index, "strides")
    if isinstance(out_backprop, PrivateTensorBase):
        with tf.device(out_backprop.owner):
            g = PModel.sum_pool_grad(input_shape, grad=out_backprop.inner_value,
                                     ksize=ksize, strides=strides, padding=padding)
            z = PrivateTensor(owner=out_backprop.owner, fixedpoint=out_backprop.fixedpoint, inner_value=g)
            return z
    elif isinstance(out_backprop, SharedPairBase):
        with tf.device(out_backprop.ownerL):
            gL = PModel.sum_pool_grad(input_shape, grad=out_backprop.xL.inner_value,
                                      ksize=ksize, strides=strides, padding=padding)
        with tf.device(out_backprop.ownerR):
            gR = PModel.sum_pool_grad(input_shape, grad=out_backprop.xR.inner_value,
                                      ksize=ksize, strides=strides, padding=padding)
        z = SharedPair(ownerL=out_backprop.ownerL,
                       ownerR=out_backprop.ownerR,
                       xL=SharedTensorBase(inner_value=gL, module=out_backprop.xL.module, shape=list(gL.shape)),
                       xR=SharedTensorBase(inner_value=gR, module=out_backprop.xR.module, shape=list(gR.shape)),
                       fixedpoint=out_backprop.fixedpoint)
        return z
    else:
        raise TypeError(
            "Expected PrivateTensorBase or SharedPairBase for 'out_backprop' argument to 'average' Op, not %r." % out_backprop)


def max_pool_div2(x: SharedPair,
                  axis, return_index=False):
    """
    :param x:  A SharedPair of shape [d1,d2,...,di,...dn]
    :param axis:  the i
    :param return_index:  bool
    :return:  if return_index=False
                 A SharedPair of shape [d1,d2,...,di/2,...dn]
            else:
                a tuple of A SharedPair of shape [d1,d2,...,di/2,...dn] and the
                    SharedPair of shape  [d1,d2,...,di/2,...dn] of module==2
    """
    xshape = x.shape
    xshape = xshape[0:axis] + [xshape[axis] // 2, 2] + xshape[axis + 1:]
    # print("xshape=", xshape)
    x = x.reshape(xshape)
    x0, x1 = x.split(size_splits=[1, 1], axis=axis + 1)
    x0 = x0.squeeze(axis=axis + 1)
    x1 = x1.squeeze(axis=axis + 1)
    # print("x0=", x0)
    # print("x1=", x1)
    s = drelu_binary(x1-x0)
    # print("s=", s)
    y = select_share(s, x1, x0)
    if return_index:
        return y, s
    else:
        return y


def max_pool_div2_back(ploss_py: SharedPair, axis, s: SharedPair):
    """
    (ploss_py,0) if s==0
    (0, ploss_py) if s==1

    :param ploss_py: SharedPair of shape [d1,...,di,...dn]
    :param axis:  i
    :param s:  SharedPair of shape [d1,...,di,...dn]
    :return:  SharedPair of shape [d1,...,2di,...dn]
    """
    if s.xL.module != 2:
        raise Exception("must have s.module==2")
    if ploss_py.shape[:axis] != s.shape[:axis]:
        raise Exception("must have ploss_py.shape[:axis]==s.shape[:axis]")
    if ploss_py.shape[axis + 1:] != s.shape[axis + 1:]:
        raise Exception("must have ploss_py.shape[axis+1:]==s.shape[axis+1:], "
                        "but ploss_py.shape={}, s.shape={}, axis+1={}".format(ploss_py.shape, s.shape, axis+1))
    # print("s=", s)
    # print("ploss_py=", ploss_py)
    new_shape = ploss_py.shape[:axis] + [-1] + ploss_py.shape[axis + 1:]
    ploss_py = ploss_py.expend_dims(axis=axis+1)
    s = s.expend_dims(axis=axis+1)
    sploss_py = select_share(s, ploss_py)
    ploss_px = (ploss_py - sploss_py).concat(sploss_py, axis=axis+1)
    ploss_px = ploss_px.reshape(new_shape)
    return ploss_px


def max_pool2d(x: Union[PrivateTensor, SharedPair],
               ksize, return_s=False):
    if len(ksize) == 2:
        ksize = [1, ksize[0], ksize[1], 1]
    if len(x.shape) != len(ksize):
        raise Exception("must have len(x.shape) == len(ksize), but x.shape={}, ksize={}".format(x.shape, ksize))
    if return_s:
        index_list = []
        for axis in range(len(ksize)):
            size_in_axis = ksize[axis]
            while size_in_axis > 1:
                x, index = max_pool_div2(x, axis, return_index=True)
                size_in_axis //= 2
                index_list.append(index)
        return x, index_list
    else:
        for axis in range(len(ksize)):
            size_in_axis = ksize[axis]
            while size_in_axis > 1:
                x = max_pool_div2(x, axis)
                size_in_axis //= 2
        return x


def max_pool2d_back(ploss_py: SharedPair, ksize, index_list=None, x: Union[PrivateTensor, SharedPair] = None):
    # print("x=", x)
    if len(ksize)==2:
        ksize = [1, ksize[0], ksize[1], 1]
    #print("ksize=", ksize)
    #print("ploss_py=", ploss_py)
    #print("index_list=", index_list)
    if index_list is None:
        y, index_list = max_pool2d(x, ksize, return_s=True)
    i = len(index_list) - 1
    ploss_px = ploss_py
    for axis in range(len(ksize) - 1, -1, -1):
        # print("axis=", axis)
        size_in_axis = ksize[axis]
        # print("size_in_axis=", size_in_axis)
        while size_in_axis > 1:
            ploss_px = max_pool_div2_back(ploss_px, axis=axis, s=index_list[i])
            i -= 1
            size_in_axis //= 2
    return ploss_px



#!/usr/bin/env python
# coding=utf-8
import numpy as np
from stensorflow.ml.nn.layers.conv2d import *
from stensorflow.ml.nn.layers.pooling import *


class Flatten(Layer):
    """
    Flattens the input. Does not affect the batch size.
    If inputs are shaped `(batch,)` without a channel dimension,
    then flattening adds an extra channel dimension and output shapes are `(batch, 1)`.
    """
    def __init__(self, output_dim, fathers: List[Layer]):
        if len(fathers) != 1:
            raise Exception("must have len(fathers) == 1 ")
        self.input_shape = fathers[0].output_dim
        # if isinstance(fathers[0], Conv2d) or isinstance(fathers[0], AveragePooling2D):
        if output_dim is None:
            output_dim = np.prod(self.input_shape[:], dtype=int)
        super(Flatten, self).__init__(output_dim=output_dim, fathers=fathers)

    def func(self, w: List[SharedVariablePair], x: List[Union[PrivateTensor, SharedPair]]):
        """
        :param w:
        :param x: input
        :return: Has the same type as `x`.
        """
        if len(x) != 1:
            raise Exception("must have len(x) == 1")
        # only support data_format is "NHWC"
        # 更新input_shape
        input_shape = x[0].shape  # return a list
        self.input_shape = input_shape  # save shape information
        flattened_dim = np.prod(input_shape[1:], dtype=int)
        # 这里进行了一次隐蔽的类型转换
        y = x[0].reshape([-1, flattened_dim])
        if isinstance(y, SharedPairBase):
            z = SharedPair(ownerL=y.ownerL, ownerR=y.ownerR,
                           xL=y.xL, xR=y.xR, fixedpoint=y.fixedpoint)
        if isinstance(y, PrivateTensorBase):
            z = PrivateTensor.from_PrivteTensorBase(y)
        return z

    def pull_back(self, w, x, y, ploss_py):
        """
        :param w:
        :param x:
        :param y:
        :param ploss_py: loss gradient Delta
        :return:
        """
        if len(x) != 1:
            raise Exception("must have len(x)==1")
        if not self.input_shape:
            raise Exception("input_shape != None")
        ploss_px = ploss_py.reshape(self.input_shape)
        if isinstance(ploss_px, SharedPairBase):
            z = SharedPair(ownerL=ploss_px.ownerL, ownerR=ploss_px.ownerR,
                           xL=ploss_px.xL, xR=ploss_px.xR, fixedpoint=ploss_px.fixedpoint)
        if isinstance(ploss_px, PrivateTensorBase):
            z = PrivateTensor.from_PrivteTensorBase(ploss_px)
        ploss_px = {self.fathers[0]: z}
        return [], ploss_px


class FlattenLocal(Layer):
    def __init__(self, output_dim, fathers: List[Layer], owner):
        if len(fathers) != 1:
            raise Exception("must have len(fathers) == 1 ")
        self.owner = owner
        self.input_shape = fathers[0].output_dim
        if isinstance(fathers[0], Conv2dLocal) or isinstance(fathers[0], AveragePooling2DLocal):
            output_dim = np.prod(self.input_shape[:], dtype=int)
        super(FlattenLocal, self).__init__(output_dim=output_dim, fathers=fathers)

    def func(self, w: List[SharedVariablePair], x: [PrivateTensor, SharedPair]):
        if len(x) != 1:
            raise Exception("must have len(x) == 1")
        x = x[0].to_private(owner=self.owner)
        x = PrivateTensor.from_PrivteTensorBase(x)
        # only support data_format is "NHWC"
        input_shape = x.shape  # return a list
        self.input_shape = input_shape  # save shape information
        flattened_dim = np.prod(input_shape[1:], dtype=int)
        # 这里进行了一次隐蔽的类型转换
        y = x.reshape([-1, flattened_dim])
        y = PrivateTensor.from_PrivteTensorBase(y)
        return y

    def pull_back(self, w, x, y, ploss_py):
        if len(x) != 1:
            raise Exception("must have len(x)==1")
        if not self.input_shape:
            raise Exception("input_shape != None")
        ploss_py = ploss_py.to_private(owner=self.owner)
        ploss_py = PrivateTensor.from_PrivteTensorBase(ploss_py)
        # 一次隐式的类型转换
        ploss_px = ploss_py.reshape(self.input_shape)
        ploss_px = PrivateTensor.from_PrivteTensorBase(ploss_px)
        # 测试单层而坐的转换
        ploss_px = {self.fathers[0]: ploss_px}
        return [], ploss_px



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : select
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-06-10 15:30
   Description : description what the main function of this file
"""

from stensorflow.basic.basic_class.pair import SharedPair
from stensorflow.basic.basic_class.private import PrivateTensor
from stensorflow.basic.basic_class.base import SharedTensorBase, SharedPairBase, PrivateTensorBase
from typing import Union
from stensorflow.basic.protocol.module_transform import module_transform
from stensorflow.basic.protocol.pm1_act import pm1_act, SharedTensorInt65, SharedPairInt65, pm1_pair_act_pair, \
    pm1_pair_act
from stensorflow.global_var import StfConfig
from stensorflow.exception.exception import StfCondException, StfEqualException, StfTypeException, StfValueException
import tensorflow as tf


def native_select(s: SharedPair, x: SharedPair, y: SharedPair = None, prf_flag=None,
                  compress_flag=None) -> SharedPair:
    """
     select using multiplication
    :param compress_flag:
    :param s: SharedPair  module=2
    :param x: SharedPai
    :param y: SharedPair, default=0
    :return: x if s else y
    """
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag

    if s.xL.module != 2:
        raise StfValueException("s.module", 2, s.xL.module)
    if s.ownerL != x.ownerL or s.ownerR != x.ownerR:
        x = x.mirror()
    if s.ownerL != x.ownerL or s.ownerR != x.ownerR:
        raise Exception("must have s.ownerL == x.ownerL and s.ownerR == x.ownerR")

    if y is not None:
        if x.xL.module != y.xL.module:
            raise Exception("must have x.module==y.module")
        if x.ownerL != y.ownerL or x.ownerR != y.ownerR:
            y = y.mirror()
        if x.ownerL != y.ownerL or x.ownerR != y.ownerR:
            raise Exception("must have x.ownerL == y.ownerL and x.ownerR == y.ownerR")
        if not (s.shape == x.shape and s.shape == y.shape):
            raise Exception("must have  s.shape==x.shape and s.shape==y.shape")
        s = module_transform(s, x.xL.module, prf_flag=prf_flag, compress_flag=compress_flag)
        s = SharedPair.from_SharedPairBase(s)
        return s * (x - y) + y
    else:
        s = module_transform(s, x.xL.module, prf_flag=prf_flag, compress_flag=compress_flag)
        s = SharedPair.from_SharedPairBase(s)
        return s * x


def _select_sharedpair_sharedpair(s: SharedPairBase, x: SharedPairBase, y: SharedPairBase = None,
                                  prf_flag=None, compress_flag=None) -> SharedPair:
    """

    :param s:
    :param x:
    :param y:
    :return:  if y==None:
                sx
            else:
                s(x-y)+y

    """
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag
    if x.xL.module is not None:
        raise StfCondException("x.xL.module is None", "x.xL.module={}".format(x.xL.module))
    if s.xL.module != 2:
        raise StfValueException("s.xL.module", 2, s.xL.module)
    if s.ownerL != x.ownerL or s.ownerR != x.ownerR:
        x = x.mirror()
    if s.ownerL != x.ownerL or s.ownerR != x.ownerR:
        raise StfEqualException("s.owner", "x.owner", (s.ownerL, s.ownerR), (x.ownerL, x.ownerR))

    if y is None:
        if x.xL.module is not None and x.xL.module % 2 == 0:
            with tf.device(x.ownerL):
                xL = SharedTensorBase(inner_value=x.xL, module=2 * x.xL.module)
            with tf.device(x.ownerR):
                xR = SharedTensorBase(inner_value=x.xR, module=2 * x.xR.module)

            x_lift = SharedPairBase(ownerL=x.ownerL, ownerR=x.ownerR, xL=xL, xR=xR, fixedpoint=x.fixedpoint)
            two_sx_lift = x_lift - pm1_pair_act_pair(s, x_lift, StfConfig.RS[0], prf_flag, compress_flag)
            with tf.device(x.ownerL):
                sxL = two_sx_lift.xL.inner_value // 2
                sxL = SharedTensorBase(inner_value=sxL, module=xL.module)

            with tf.device(x.ownerR):
                sxR = two_sx_lift.xR.inner_value - two_sx_lift.xR.inner_value // 2
                sxR = SharedTensorBase(inner_value=sxR, module=xR.module)

            sx = SharedPairBase(ownerL=x.ownerL, ownerR=x.ownerR, xL=sxL, xR=sxR, fixedpoint=x.fixedpoint)

        elif x.xL.module is not None:
            inv_2 = x.xL.module - (x.xL.module // 2)
            two_sx = x - pm1_pair_act_pair(s, x, StfConfig.RS[0])
            sx = inv_2 * two_sx
        else:
            lift_x = SharedPairInt65.from_SharedPair(x)
            s_act_lift_x = pm1_pair_act_pair(s, lift_x, StfConfig.RS[0], prf_flag, compress_flag)
            two_sx = lift_x - s_act_lift_x
            sx = two_sx.half_of_even()
            sx = SharedPair.from_SharedPairBase(sx)
        return sx

    else:
        if x.xL.module != y.xL.module:
            raise StfEqualException("x.xL.module", "y.xL.module", x.xL.module, y.xL.module)
        if x.ownerL != y.ownerL or x.ownerR != y.ownerR:
            y = y.mirror()
        if x.ownerL != y.ownerL or x.ownerR != y.ownerR:
            raise StfEqualException("x.owner", "y.owner", (x.ownerL, x.ownerR), (y.ownerL, y.ownerR))
        if not (s.shape == x.shape and s.shape == y.shape):
            raise StfCondException("s.shape == x.shape and s.shape == y.shape",
                                   "s.shape={}, x.shape={}, y.shape={}".format(s.shape, x.shape, y.shape))
        z = x - y
        return _select_sharedpair_sharedpair(s, z) + y


def _select_private_private(s: PrivateTensor, x: PrivateTensor, y: PrivateTensor = None,
                            prf_flag=None, compress_flag=None):
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag

    if s.module != 2:
        raise StfValueException("s.module", 2, s.module)
    if x.module is not None:
        raise StfCondException("x.module is None", "x.module={}".format(x.module))
    if y is None:
        if s.owner == x.owner:
            with tf.device(s.owner):
                s_mt = s.identity()
                s_mt.module = x.module
            return x * s_mt
        else:
            with tf.device(x.owner):
                lift_x = SharedTensorInt65.load_from_PrivateTensor(x)
            s_act_list_x = pm1_act(s.to_SharedTensor(), lift_x, s.owner, x.owner, RS_owner=StfConfig.RS[0],
                                   prf_flag=prf_flag, compress_flag=compress_flag)
            two_sx = SharedPairInt65.load_from_SharedTensorInt65(lift_x, owner=x.owner,
                                                                 other_owner=s.owner) - s_act_list_x
            sx = two_sx.half_of_even()
            sx.fixedpoint = x.fixedpoint
            return sx
    else:
        if x.owner == y.owner:
            z = _select_private_private(s=s, x=x - y)
        else:
            z = _select_private_sharedpair(s=s, x=x - y)
        if isinstance(z, SharedPairBase) and not isinstance(z, SharedPair):
            z = SharedPair.from_SharedPairBase(z)
        elif isinstance(z, PrivateTensorBase) and not isinstance(z, PrivateTensor):
            z = PrivateTensor.from_PrivteTensorBase(z)
        return z + y


def _select_private_sharedpair(s: PrivateTensor, x: SharedPair, y: SharedPair = None,
                               prf_flag=None, compress_flag=None):
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag

    if s.module != 2:
        raise StfValueException("s.module", 2, s.module)
    if s.shape != x.shape:
        raise StfEqualException("s.shape", "x.shape", s.shape, x.shape)
    if x.xL.module is not None:
        raise StfCondException("x.xL.module is None", "x.xL.module={}".format(x.xL.module))
    if y is None:
        xL = PrivateTensor(owner=x.ownerL, inner_value=x.xL.inner_value, fixedpoint=x.fixedpoint, module=x.xL.module)
        xR = PrivateTensor(owner=x.ownerR, inner_value=x.xR.inner_value, fixedpoint=x.fixedpoint, module=x.xR.module)
        zL = _select_private_private(s=s, x=xL, prf_flag=prf_flag, compress_flag=compress_flag)
        zR = _select_private_private(s=s, x=xR, prf_flag=prf_flag, compress_flag=compress_flag)
        return zL + zR
    else:
        if x.xL.module != y.xL.module:
            raise StfEqualException("x.xL.module", "y.xL.module", x.xL.module, y.xL.module)
        if x.ownerL != y.ownerL or x.ownerR != y.ownerR:
            y = y.mirror()
        if x.ownerL != y.ownerL or x.ownerR != y.ownerR:
            raise StfEqualException("x.owner", "y.owner", (x.ownerL, x.ownerR), (y.ownerL, y.ownerR))
        if not (s.shape == x.shape and s.shape == y.shape):
            raise StfCondException("s.shape == x.shape and s.shape == y.shape",
                                   "s.shape={}, x.shape={}, y.shape={}".format(s.shape, x.shape, y.shape))
        z = x - y
        return _select_private_sharedpair(s, z) + y


def _select_sharedpair_private(s: SharedPairBase, x: PrivateTensorBase, y: PrivateTensorBase = None,
                               prf_flag=None, compress_flag=None):
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag
    if x.module is not None:
        raise StfCondException("x.module is None", "x.module={}".format(x.module))
    if s.xL.module != 2:
        raise StfValueException("s.xL.module", 2, s.xL.module)
    if s.shape != x.shape:
        raise StfEqualException("s.shape", "x.shape", s.shape, x.shape)
    if y is None:
        with tf.device(x.owner):
            lift_x = SharedTensorInt65.load_from_PrivateTensor(x)
        s_act_lift_x = pm1_pair_act(s, lift_x, x_owner=x.owner, RS_owner=StfConfig.RS[0],
                                    prf_flag=prf_flag, compress_flag=compress_flag)
        if s.ownerL == x.owner:
            other_owner = s.ownerR
        elif s.ownerR == x.owner:
            other_owner = s.ownerL
        else:
            raise StfCondException("x.owner==s.ownerL or x.owner==s.ownerR",
                                   "x.owner={}, s.ownerL={}, s.ownerR={}".format(x.owner, s.ownerL, s.ownerR))

        two_sx = SharedPairInt65.load_from_SharedTensorInt65(lift_x, owner=x.owner,
                                                             other_owner=other_owner) - s_act_lift_x
        sx = two_sx.half_of_even()
        sx.fixedpoint = x.fixedpoint
        return sx
    else:
        if x.module != y.module:
            raise StfEqualException("x.module", "y.module", x.module, y.module)
        if x.shape != y.shape:
            raise StfEqualException("x.shape", "y.shape", x.shape, y.shape)
        if x.owner == y.owner:
            z = x - y
            return _select_sharedpair_private(s, z) + y
        else:
            z = x - y
            return _select_sharedpair_sharedpair(s, z) + y


def select_share(s: Union[PrivateTensorBase, SharedPairBase, tf.Tensor], x: Union[PrivateTensorBase, SharedPairBase],
                 y: Union[PrivateTensorBase, SharedPairBase] = 0, prf_flag=None, compress_flag=None):
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag
    if y == 0:
        y = None

    if isinstance(s, SharedPairBase) and isinstance(x, SharedPairBase):
        return _select_sharedpair_sharedpair(s, x, y, prf_flag, compress_flag)
    elif isinstance(s, SharedPairBase) and isinstance(x, PrivateTensorBase):
        return _select_sharedpair_private(s, x, y, prf_flag, compress_flag)
    elif isinstance(s, PrivateTensorBase) and isinstance(x, PrivateTensorBase):
        return _select_private_private(s, x, y, prf_flag, compress_flag)
    elif isinstance(s, PrivateTensorBase) and isinstance(x, SharedPairBase):
        return _select_private_sharedpair(s, x, y, prf_flag, compress_flag)
    elif isinstance(s, tf.Tensor):
        if y is None:
            return s * x
        else:
            return s * (x - y) + y
    else:
        raise StfTypeException("s,x", "PrivateTensorBase or SharedPairBase", [type(s), type(x)])



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : DNN_with_SL
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-12-17 11:27
   Description : description what the main function of this file
"""

from stensorflow.ml.nn.layers.layer import Layer
from stensorflow.exception.exception import StfValueException
from stensorflow.ml.nn.networks.NN import NN
import tensorflow as tf
from typing import Union, List
from stensorflow.random.random import random_init
from stensorflow.basic.basic_class.private import PrivateTensor
from stensorflow.basic.basic_class.pair import SharedPair
from stensorflow.ml.nn.layers.input import Input
from stensorflow.ml.nn.layers.dense import Dense, Dense_Local
from stensorflow.ml.nn.layers.sum import Sum
from stensorflow.ml.nn.layers.loss import Loss, BinaryCrossEntropyLossWithSigmoid, \
    BinaryCrossEntropyLossWithSigmoidLocal, CrossEntropyLossWithSoftmax, CrossEntropyLossWithSoftmaxLocal
from stensorflow.ml.nn.layers.relu import ReLU, ReLU_Local


class DNN_with_SL(NN):
    def __init__(self, feature: PrivateTensor, label: Union[PrivateTensor, SharedPair], dense_dims: List[int],
                 secure_config: list = None, loss=None):
        """
        单方特征DNN
        :param feature:
        :param label:
        :param dense_dims: [d0, d1, ..., dn]  where d0=dim(feature) dn=dim(output)
        :param secure_config:  [n0, n1, n2]
        where n0>=1, n1,n2>=0 and n0+n1+n2=n
        it meanse secure_levels=["local"]xn0+["share"]xn1+["local"]xn2
        """

        if len(dense_dims) < 2:
            raise Exception("must have len(dense_dims)>=2")
        if secure_config is None:
            secure_config = [1, len(dense_dims) - 1, 0]

        if not isinstance(secure_config, list):
            raise Exception("secure_config must be a list of int")
        elif len(secure_config) != 3:
            raise Exception("must have len(secure_config)==3")
        elif not (
                isinstance(secure_config[0], int) and isinstance(secure_config[1], int) and isinstance(secure_config[2],
                                                                                                       int)):
            raise Exception(
                "must have isinstance(secure_config[0], int) and isinstance(secure_config[1], int) and isinstance("
                "secure_config[2], int)")
        elif secure_config[0] < 1 or secure_config[1] < 0 or secure_config[2] < 0 or sum(secure_config) != len(
                dense_dims):
            raise Exception(
                "must have secure_config[0]>=1 and secure_config[1]>=0 and secure_config[2]>=0 and sum("
                "secure_config)==len(dense_dims)")
        else:
            pass

        super(DNN_with_SL, self).__init__()
        self.secure_config = secure_config
        layer = Input(dim=feature.shape[1], x=feature)

        local_layer_owner = layer.owner
        self.addLayer(ly=layer)

        for i in range(1, secure_config[0]):
            layer = Dense_Local(output_dim=dense_dims[i], fathers=[layer], owner=local_layer_owner, with_b=True)

            self.addLayer(ly=layer)
            if i < len(dense_dims) - 1:
                layer = ReLU_Local(output_dim=dense_dims[i], fathers=[layer], owner=local_layer_owner)

                self.addLayer(ly=layer)

        for i in range(secure_config[0], secure_config[0] + secure_config[1]):
            layer = Dense(output_dim=dense_dims[i], fathers=[layer])

            self.addLayer(ly=layer)
            if i < len(dense_dims) - 1:
                layer = ReLU(output_dim=dense_dims[i], fathers=[layer])

                self.addLayer(ly=layer)

        for i in range(secure_config[0] + secure_config[1], sum(secure_config)):
            layer = Dense_Local(output_dim=dense_dims[i], fathers=[layer], owner=label.owner, with_b=True)

            self.addLayer(ly=layer)
            if i < len(dense_dims) - 1:
                layer = ReLU_Local(output_dim=dense_dims[i], fathers=[layer], owner=label.owner)

                self.addLayer(ly=layer)

        # layer_label = Input(dim=1, x=label)
        layer_label = Input(dim=dense_dims[-1], x=label)
        self.addLayer(ly=layer_label)

        if loss is None:
            loss = "BinaryCrossEntropyLossWithSigmoid"

        if loss == "BinaryCrossEntropyLossWithSigmoid" or loss == Loss.BinaryCrossEntropyLossWithSigmoid:
            if secure_config[-1] == 0:
                layer_loss = BinaryCrossEntropyLossWithSigmoid(layer_score=layer, layer_label=layer_label)
            else:
                layer_loss = BinaryCrossEntropyLossWithSigmoidLocal(layer_score=layer, layer_label=layer_label,
                                                                    owner=layer_label.owner)
        elif loss == "CrossEntropyLossWithSoftmax" or loss == Loss.CrossEntropyLossWithSoftmax:
            if secure_config[-1] == 0:
                layer_loss = CrossEntropyLossWithSoftmax(layer_score=layer, layer_label=layer_label)
            else:
                layer_loss = CrossEntropyLossWithSoftmaxLocal(layer_score=layer, layer_label=layer_label,
                                                              owner=layer_label.owner)
        else:
            raise StfValueException("loss", "BinaryCrossEntropyLossWithSigmoid or CrossEntropyLossWithSoftmax", loss)

        self.addLayer(ly=layer_loss)

    def predict(self, x, x_another=None, out_prob=True) -> Union[SharedPair, PrivateTensor]:
        self.cut_off()
        l_input = self.layers[0]
        if not isinstance(l_input, Input):
            raise Exception("l_input mast be a Input layer")
        l_input.replace(x)
        self.layers[0] = l_input

        if x_another is not None:
            l_input_another = self.layers[1]
            if not isinstance(l_input_another, Input):
                raise Exception("l_input_another mast be a Input layer")
            l_input_another.replace(x_another)
            self.layers[1] = l_input_another

        ly = self.layers[-1]
        if not isinstance(ly, Layer):
            raise Exception("l must be a Layer")
        else:
            ly.forward()

        if out_prob:
            return ly.y
        else:
            return ly.score

    def predict_to_file(self, sess, x, predict_file_name,
                        batch_num, idx, model_file_machine, record_num_ceil_mod_batch_size,
                        x_another=None, with_sigmoid=True):
        y_pred = self.predict(x=x, x_another=x_another, out_prob=with_sigmoid)

        id_y_pred = y_pred.to_tf_str(owner=model_file_machine, id_col=idx)
        sess.run(tf.compat.v1.global_variables_initializer())
        with open(predict_file_name, "w") as f:
            for batch in range(batch_num - 1):
                records = sess.run(id_y_pred)
                records = "\n".join(records.astype('str'))
                f.write(records + "\n")

            records = sess.run(id_y_pred)[0:record_num_ceil_mod_batch_size]
            records = "\n".join(records.astype('str'))
            f.write(records + "\n")

    def print(self, sess, model_file_machine="R"):
        for ly in self.layers:
            if isinstance(ly, Dense) or isinstance(ly, Dense_Local):
                for weight in ly.w:
                    weight = weight.to_tf_tensor(owner=model_file_machine)
                    sess.run(weight)

    def save(self, sess, model_file_machine="R", path="./output/model"):
        print("save model...")
        i = 0
        for ly in self.layers:
            if isinstance(ly, Dense) or isinstance(ly, Dense_Local):
                ly.save(model_file_machine, sess, path + "/param_{}".format(i))
            i += 1

    def load(self, path="./output/model/"):
        print("load model...")
        i = 0
        for ly in self.layers:
            if isinstance(ly, Dense) or isinstance(ly, Dense_Local):
                ly.load(path + "/param_{}".format(i))
            i += 1


class DNN_with_SL2(NN):
    def __init__(self, feature: PrivateTensor, feature_another: PrivateTensor, label: PrivateTensor,
                 dense_dims: List[int], secure_config: list = None, loss=None):
        """

        :param feature:
        :param label:
        :param dense_dims: [d0, d1, ..., dn]  where d0=dim(feature) dn=dim(output)
        :param feature_another:
        :param secure_config:  [n0, n1, n2]
        where n0>=1, n1,n2>=0 and n0+n1+n2=n+1
        it meanse secure_levels=["local"]xn0+["share"]xn1+["local"]xn2
        """

        if len(dense_dims) < 2:
            raise Exception("must have len(dense_dims)>=2")
        if secure_config is None:
            secure_config = [1, len(dense_dims) - 1, 0]

        if not isinstance(secure_config, list):
            raise Exception("secure_config must be a list of int")
        elif len(secure_config) != 3:
            raise Exception("must have len(secure_config)==3")
        elif not (
                isinstance(secure_config[0], int) and isinstance(secure_config[1], int) and isinstance(secure_config[2],
                                                                                                       int)):
            raise Exception(
                "must have isinstance(secure_config[0], int) and isinstance(secure_config[1], int) and isinstance("
                "secure_config[2], int)")
        elif secure_config[0] < 1 or secure_config[1] < 0 or secure_config[2] < 0 or sum(secure_config) != len(
                dense_dims):
            raise Exception(
                "must have secure_config[0]>=1 and secure_config[1]>=0 and secure_config[2]>=0 and sum("
                "secure_config)==len(dense_dims)")
        else:
            pass

        super(DNN_with_SL2, self).__init__()
        self.secure_config = secure_config
        layer = Input(dim=feature.shape[1], x=feature)
        self.addLayer(ly=layer)

        if feature.shape[1] + feature_another.shape[1] != dense_dims[0]:
            raise Exception("must have feature.shape[1]+feature_another.shape[1] == dense_dims[0]")
        if feature.owner == feature_another.owner:
            raise Exception("must have feature.owner != feature_another.owner")
        layer_another = Input(dim=feature_another.shape[1], x=feature_another)
        self.addLayer(layer_another)

        if secure_config[0] == 1:  # local, share, .....
            layer = Dense(output_dim=dense_dims[1], fathers=[layer, layer_another])
            self.addLayer(ly=layer)
        elif secure_config[0] == 2:  # local, local, share, ...
            layer = Dense_Local(output_dim=dense_dims[1], fathers=[layer], owner=layer.owner, with_b=True)
            self.addLayer(ly=layer)
            layer_another = Dense_Local(output_dim=dense_dims[1], fathers=[layer_another], owner=layer_another.owner,
                                        with_b=False)
            self.addLayer(ly=layer_another)
            layer = Sum(output_dim=dense_dims[1], fathers=[layer, layer_another])
            self.addLayer(ly=layer)
        else:
            raise Exception("must have secure_config[0]==1 or 2")
        if len(dense_dims) > 2:
            layer = ReLU(output_dim=dense_dims[1], fathers=[layer])
            self.addLayer(ly=layer)

        for i in range(2, secure_config[0] + secure_config[1]):
            layer = Dense(output_dim=dense_dims[i], fathers=[layer])
            self.addLayer(ly=layer)
            if len(dense_dims) > 1 + i:
                layer = ReLU(output_dim=dense_dims[i], fathers=[layer])
                self.addLayer(ly=layer)

        for i in range(secure_config[0] + secure_config[1], sum(secure_config)):
            layer = Dense_Local(output_dim=dense_dims[i], fathers=[layer], owner=label.owner, with_b=True)
            self.addLayer(ly=layer)
            if len(dense_dims) > 1 + i:
                layer = ReLU_Local(output_dim=dense_dims[i], fathers=[layer], owner=label.owner)
                self.addLayer(ly=layer)

        layer_label = Input(dim=1, x=label)
        self.addLayer(ly=layer_label)

        if loss is None:
            loss = "BinaryCrossEntropyLossWithSigmoid"

        if loss == "BinaryCrossEntropyLossWithSigmoid" or loss == Loss.BinaryCrossEntropyLossWithSigmoid:
            if secure_config[-1] == 0:
                layer_loss = BinaryCrossEntropyLossWithSigmoid(layer_score=layer, layer_label=layer_label)
            else:
                layer_loss = BinaryCrossEntropyLossWithSigmoidLocal(layer_score=layer, layer_label=layer_label,
                                                                    owner=layer_label.owner)
        elif loss == "CrossEntropyLossWithSoftmax" or loss == Loss.CrossEntropyLossWithSoftmax:
            if secure_config[-1] == 0:
                layer_loss = CrossEntropyLossWithSoftmax(layer_score=layer, layer_label=layer_label)
            else:
                layer_loss = CrossEntropyLossWithSoftmaxLocal(layer_score=layer, layer_label=layer_label,
                                                              owner=layer_label.owner)
        else:
            raise StfValueException("loss", "BinaryCrossEntropyLossWithSigmoid or CrossEntropyLossWithSoftmax", loss)

        self.addLayer(ly=layer_loss)

    def predict(self, x, x_another=None, with_sigmoid=True) -> Union[SharedPair, PrivateTensor]:
        self.cut_off()
        l_input = self.layers[0]
        if not isinstance(l_input, Input):
            raise Exception("l_input mast be a Input layer")
        l_input.replace(x)
        # self.layers[0] = l_input

        if x_another is not None:
            l_input_another = self.layers[1]
            if not isinstance(l_input_another, Input):
                raise Exception("l_input_another mast be a Input layer")
            l_input_another.replace(x_another)
            # self.layers[1] = l_input_another

        ly = self.layers[-1]
        if not isinstance(ly, Layer):
            raise Exception("l must be a Layer")
        else:
            ly.forward()

        if with_sigmoid:
            return ly.y
        else:
            return ly.score

    def predict_to_file(self, sess, x, predict_file_name,
                        batch_num, idx, model_file_machine, record_num_ceil_mod_batch_size,
                        x_another=None, with_sigmoid=True):
        y_pred = self.predict(x=x, x_another=x_another, with_sigmoid=with_sigmoid)

        id_y_pred = y_pred.to_tf_str(owner=model_file_machine, id_col=idx)
        random_init(sess)
        with open(predict_file_name, "w") as f:
            for batch in range(batch_num - 1):
                records = sess.run(id_y_pred)
                records = "\n".join(records.astype('str'))
                f.write(records + "\n")

            records = sess.run(id_y_pred)[0:record_num_ceil_mod_batch_size]
            records = "\n".join(records.astype('str'))
            f.write(records + "\n")

    def print(self, sess, model_file_machine="R"):
        for ly in self.layers:
            if isinstance(ly, Dense) or isinstance(ly, Dense_Local):
                for weight in ly.w:
                    weight_tf = weight.to_tf_tensor(owner=model_file_machine)
                    print(sess.run(weight_tf))

    def save(self, sess, model_file_machine="R", path="./output/model"):
        print("save model...")
        i = 0
        for ly in self.layers:
            if isinstance(ly, (Dense, Dense_Local)):
                ly.save(model_file_machine, sess, path + "/param_{}".format(i))
            i += 1

    def load(self, path="./output/model/"):
        print("load model...")
        i = 0
        for ly in self.layers:
            if isinstance(ly, (Dense, Dense_Local)):
                ly.load(path + "/param_{}".format(i))
            i += 1



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2021 All Rights Reserved.
   ------------------------------------------------------
   File Name : homo_mat_mul
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2021/9/30 下午5:31
   Description : description what the main function of this file
"""

from stensorflow.basic.basic_class.base import PrivateTensorBase, SharedPairBase, SharedTensorBase
from stensorflow.exception.exception import StfValueException, StfEqualException, StfCondException
import tensorflow as tf
from typing import Union
from stensorflow.basic.operator.algebra import stack
from stensorflow.global_var import StfConfig
from stensorflow.basic.protocol.bilinear_map import BM_PrivateTensor_SharedPair, BM_PrivateTensor_PrivateTensor, \
    BM_SharedPair_SharedPair, BM_SharedPair_PrivateTensor
from stensorflow.homo_enc.homo_enc import gene_key, enc, mat_mul_vec_to_share, dec


def matmulvec_homo(mat: Union[PrivateTensorBase, SharedPairBase], x: Union[PrivateTensorBase, SharedPairBase],
                   fixed_point=None) -> SharedPairBase:
    mat_is_private = isinstance(mat, PrivateTensorBase)
    x_is_private = isinstance(x, PrivateTensorBase)

    if mat_is_private and not x_is_private:
        return matmulvec_privat_pair(mat, x, fixed_point)
    elif mat_is_private and x_is_private:
        return matmulvec_privat_private(mat, x, fixed_point)
    else:
        raise NotImplementedError


def matmulvec_privat_private(mat: PrivateTensorBase, x: PrivateTensorBase, fixed_point=None) -> SharedPairBase:
    if isinstance(mat, PrivateTensorBase) and isinstance(x, PrivateTensorBase):
        pass
    else:
        raise NotImplementedError
    if mat.shape[1] != x.shape[0]:
        raise StfEqualException("mat.shape[1]", "x.shape[0]", mat.shape[1], x.shape[0])
    if mat.owner == x.owner:
        return PrivateTensorBase.__matmul__(mat, x, fixed_point)
    else:
        pass
    row_num = mat.shape[0]
    with tf.device(x.owner):
        sk, pk, gk = gene_key()
        cipher_vec_x = enc(pk, x.inner_value)
    with tf.device(mat.owner):
        plaintext_out_mat, cipher_out_vec = mat_mul_vec_to_share(pk, gk, mat.inner_value,
                                                                 cipher_vec_x)
        plaintext_out_mat = tf.reshape(plaintext_out_mat, [row_num])
    with tf.device(x.owner):
        plaintext_out_vec = dec(sk, [row_num], cipher_out_vec)
        plaintext_out_vec = tf.reshape(plaintext_out_vec, [row_num])

    with tf.device(mat.owner):
        xL = SharedTensorBase(inner_value=plaintext_out_mat)
    with tf.device(x.owner):
        xR = SharedTensorBase(inner_value=plaintext_out_vec)
    r = SharedPairBase(ownerL=mat.owner, ownerR=x.owner, xL=xL, xR=xR,
                       fixedpoint=mat.fixedpoint + x.fixedpoint)
    if fixed_point is None:
        return r
    else:
        return r.dup_with_precision(new_fixedpoint=fixed_point)


def matmulvec_privat_pair(mat: PrivateTensorBase, x: SharedPairBase, fixed_point=None) -> SharedPairBase:
    if isinstance(mat, PrivateTensorBase) and isinstance(x, SharedPairBase):
        pass
    else:
        raise NotImplementedError
    if mat.shape[1] != x.shape[0]:
        raise StfEqualException("mat.shape[1]", "x.shape[0]", mat.shape[1], x.shape[0])
    row_num = mat.shape[0]
    if x.ownerL == mat.owner:
        vec_owner = x.ownerR
        vec_x = x.xR.inner_value
    elif x.ownerR == mat.owner:
        vec_owner = x.ownerL
        vec_x = x.xL.inner_value
    else:
        raise StfValueException("mat.owner", "x.ownerL or x.ownerR", mat.owner)

    with tf.device(vec_owner):
        homo_module = StfConfig.homo_module
        sk, pk, gk = homo_module.gen_key(tf.Variable(initial_value=[0]))
        cipher_vec_x = homo_module.enc(pk, vec_x)
    with tf.device(mat.owner):
        plaintext_out_mat, cipher_out_vec = homo_module.mat_mul_vec_to_share(pk, gk, mat.inner_value, cipher_vec_x)
        plaintext_out_mat = tf.reshape(plaintext_out_mat, [mat.shape[0]])
    with tf.device(vec_owner):
        plaintext_out_vec = homo_module.dec(sk, [row_num], cipher_out_vec)
        plaintext_out_vec = tf.reshape(plaintext_out_vec, [mat.shape[0]])

    if x.ownerL == mat.owner:
        with tf.device(x.ownerL):
            mx_local = tf.matmul(mat.inner_value, x.xL.inner_value)
            mx_local = tf.reshape(mx_local, plaintext_out_mat.shape)
            xL = SharedTensorBase(inner_value=plaintext_out_mat + mx_local)
        with tf.device(x.ownerR):
            xR = SharedTensorBase(inner_value=plaintext_out_vec)
    elif x.ownerR == mat.owner:
        with tf.device(x.ownerL):
            xL = SharedTensorBase(inner_value=plaintext_out_vec)
        with tf.device(x.ownerR):
            mx_local = tf.matmul(mat.inner_value, x.xR.inner_value)
            mx_local = tf.reshape(mx_local, plaintext_out_mat.shape)
            xR = SharedTensorBase(inner_value=plaintext_out_mat + mx_local)
    else:
        raise StfValueException("mat.owner", "x.ownerL or x.ownerR", mat.owner)
    r = SharedPairBase(ownerL=x.ownerL, ownerR=x.ownerR, xL=xL, xR=xR,
                       fixedpoint=mat.fixedpoint + x.fixedpoint)
    if fixed_point is None:
        return r
    else:
        return r.dup_with_precision(new_fixedpoint=fixed_point)


def matmul_homo_dim2(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase],
                     fixed_point=None) -> SharedPairBase:
    if len(x.shape) != 2:
        raise StfValueException("len(x.shape)", 2, len(x.shape))
    if len(y.shape) != 2:
        raise StfValueException("len(y.shape)", 2, len(y.shape))
    if x.shape[1] != y.shape[0]:
        raise StfEqualException("x.shape[1]", "y.shape[0]", x.shape[1], y.shape[0])
    if y.shape[1] == 1:
        z = matmulvec_homo(x, y, fixed_point=fixed_point).reshape([-1, 1])
        # print("z=",z)
    else:
        y_columns = y.split(size_splits=y.shape[1], axis=1)
        z_list = []
        for yj in y_columns:
            z_list.append(matmulvec_homo(x, yj, fixed_point=fixed_point))
        # print("z_list[0]=", z_list[0])
        z = stack(z_list, axis=1)
        print("z=", z)
    return z


def matmul_homo(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase],
                fixed_point=None) -> SharedPairBase:
    if len(x.shape) == 2 and len(y.shape) == 2:
        return matmul_homo_dim2(x, y, fixed_point)
    else:
        if len(x.shape) != len(y.shape):
            raise StfEqualException("len(x.shape)", "len(y.shape)", len(x.shape), len(y.shape))
        m = x.shape[-2]
        n = x.shape[-1]
        if x.shape[-1] != y.shape[-2]:
            raise StfEqualException("x.shape[-1]", "y.shape[-2]", x.shape[-1], y.shape[-2])
        p = y.shape[-1]
        reshaped_x = x.reshape([-1, m, n])
        reshaped_y = y.reshape([-1, n, p])
        # print("reshaped_x.shape=",reshaped_x.shape)
        # print("reshaped_y.shape=", reshaped_y.shape)
        split_x = reshaped_x.split(size_splits=reshaped_x.shape[0], axis=0)
        split_y = reshaped_y.split(size_splits=reshaped_y.shape[0], axis=0)
        list_reshape_z = []
        for (xi, yi) in zip(split_x, split_y):
            list_reshape_z.append(matmul_homo_dim2(xi.squeeze(axis=0), yi.squeeze(axis=0), fixed_point=fixed_point))
        reshape_z = stack(list_reshape_z, axis=0)
        z = reshape_z.reshape(x.shape[:-2] + [m, p])
        return z


def matmul_homo_offline(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase]):
    # print("matmul_homo_offline")
    x.serialize()
    y.serialize()
    if len(x.shape) == 2 and len(y.shape) == 2:
        z = matmul_homo_dim2(x, y)
    else:
        if len(x.shape) != len(y.shape):
            raise StfEqualException("len(x.shape)", "len(y.shape)", len(x.shape), len(y.shape))
        m = x.shape[-2]
        n = x.shape[-1]
        if x.shape[-1] != y.shape[-2]:
            raise StfEqualException("x.shape[-1]", "y.shape[-2]", x.shape[-1], y.shape[-2])
        p = y.shape[-1]
        reshaped_x = x.reshape([-1, m, n])
        reshaped_y = y.reshape([-1, n, p])
        # print("reshaped_x.shape=",reshaped_x.shape)
        # print("reshaped_y.shape=", reshaped_y.shape)
        split_x = reshaped_x.split(size_splits=reshaped_x.shape[0], axis=0)
        split_y = reshaped_y.split(size_splits=reshaped_y.shape[0], axis=0)
        list_reshape_z = []
        for (xi, yi) in zip(split_x, split_y):
            list_reshape_z.append(matmul_homo_dim2(xi.squeeze(axis=0), yi.squeeze(axis=0)))
        reshape_z = stack(list_reshape_z, axis=0)
        z = reshape_z.reshape(x.shape[:-2] + [m, p])
    z.serialize()
    return z


def matmul_homo_online(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase],
                       x_adjoint: Union[PrivateTensorBase, SharedPairBase],
                       y_adjoint: Union[PrivateTensorBase, SharedPairBase],
                       z_adjoint: Union[PrivateTensorBase, SharedPairBase], fixed_point=None):
    x_is_private = isinstance(x, PrivateTensorBase)
    x_is_shared = isinstance(x, SharedPairBase)
    y_is_private = isinstance(y, PrivateTensorBase)
    y_is_shared = isinstance(y, SharedPairBase)

    if x_is_private and y_is_private:
        assert isinstance(x_adjoint, PrivateTensorBase) and isinstance(y_adjoint, PrivateTensorBase)
        z = BM_PrivateTensor_PrivateTensor(x, y, lambda _x, _y: SharedTensorBase.__matmul__(_x, _y),
                                           prf_flag=False, x_adjoint=x_adjoint.to_SharedTensor(),
                                           y_adjoint=y_adjoint.to_SharedTensor(),
                                           u0=z_adjoint.xL, u1=z_adjoint.xR)
    elif x_is_private and y_is_shared:
        assert isinstance(x_adjoint, PrivateTensorBase) and isinstance(y_adjoint, PrivateTensorBase)
        # print("line 481 x_adjoint=", x_adjoint)
        # print("linne 482 y_adjoint=", y_adjoint)
        # print("line 483 z_adjoint.xL=", z_adjoint.xL.inner_value)
        # print("line 484 z_adjoint.xR=", z_adjoint.xR.inner_value)
        z = BM_PrivateTensor_SharedPair(x, y, lambda _x, _y: SharedTensorBase.__matmul__(_x, _y),
                                        prf_flag=False, x_adjoint=x_adjoint.to_SharedTensor(),
                                        y_adjoint=y_adjoint.to_SharedTensor(), u0=z_adjoint.xL, u1=z_adjoint.xR
                                        )
    elif x_is_shared and y_is_private:
        assert isinstance(x_adjoint, PrivateTensorBase) and isinstance(y_adjoint, PrivateTensorBase)
        z = BM_SharedPair_PrivateTensor(x, y, lambda _x, _y: SharedTensorBase.__matmul__(_x, _y),
                                        prf_flag=False, x_adjoint=x_adjoint.to_SharedTensor(),
                                        y_adjoint=y_adjoint.to_SharedTensor(),
                                        u0=z_adjoint.xL, u1=z_adjoint.xR)
    elif x_is_shared and y_is_shared:
        assert isinstance(x_adjoint, SharedPairBase) and isinstance(y_adjoint, SharedPairBase)
        z = BM_SharedPair_SharedPair(x, y, lambda _x, _y: SharedTensorBase.__matmul__(_x, _y),
                                     prf_flag=False, xL_adjoint=x_adjoint.xL, xR_adjoin=x_adjoint.xR,
                                     yL_adjoint=y_adjoint.xL, yR_adjoint=y_adjoint.xR,
                                     uL=z_adjoint.xL, uR=z_adjoint.xR)
    else:
        raise StfCondException("x, y are PrivateTensor or SharedPair", "x is, y is".format(type(x), type(y)))
    return z



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2021 All Rights Reserved.
   ------------------------------------------------------
   File Name : homo_enc.py
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2021/9/29 下午3:19
   Description : description what the main function of this file
"""
from stensorflow.global_var import StfConfig
import tensorflow as tf


def homo_init(sess=None):
    init_coll = tf.compat.v1.get_collection(StfConfig.coll_name_vars_homo)
    init_homo_op = tf.compat.v1.initialize_variables(init_coll)
    if sess is not None:
        sess.run(init_homo_op)
    else:
        return init_homo_op


def gene_key():
    homo_module = StfConfig.homo_module
    gene_key_zero = tf.Variable(initial_value=[0], trainable=False)
    tf.compat.v1.add_to_collection(StfConfig.coll_name_vars_homo, gene_key_zero)
    sk, pk, gk = homo_module.gen_key(gene_key_zero)
    return sk, pk, gk


def enc(pk, x):
    homo_module = StfConfig.homo_module
    return homo_module.enc(pk, x)


def mat_mul_vec_to_share(pk, gk, mat, cipher_vec):
    homo_module = StfConfig.homo_module
    return homo_module.mat_mul_vec_to_share(pk, gk, mat, cipher_vec)


def dec(sk, size, cipher_vec):
    homo_module = StfConfig.homo_module
    z = homo_module.dec(sk, size, cipher_vec)
    return tf.reshape(z, size)


def vec_mul_vec(pk, plain_vec, cipher_vec):
    homo_module = StfConfig.homo_module
    return homo_module.vec_mul_vec(pk, plain_vec, cipher_vec)


def cipher_to_share(size, pk, cipher):
    homo_module = StfConfig.homo_module
    cipher_out, share_out = homo_module.cipher_to_share(size, pk, cipher)
    share_out = tf.reshape(share_out, [size])
    return cipher_out, share_out


def vec_mul_vec_to_share(pk, plain_vec, cipher_vec):
    cipher_out = vec_mul_vec(pk, plain_vec, cipher_vec)
    size = tf.cast(tf.shape(plain_vec)[0], "int64")
    z = cipher_to_share(size, pk, cipher_out)
    return z



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : base
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-05-14 19:58
   Description : description what the main function of this file
"""
from stensorflow.basic.basic_class.base import PrivateTensorBase, SharedPairBase, SharedTensorBase
from stensorflow.exception.exception import StfTypeException
import tensorflow as tf
from stensorflow.global_var import StfConfig
from stensorflow.random.random import get_seed


def BM_PrivateTensor_PrivateTensor(x: PrivateTensorBase, y: PrivateTensorBase, f, prf_flag=None,
                                   x_adjoint: SharedTensorBase=None, y_adjoint: SharedTensorBase=None,
                                   u0: SharedTensorBase=None, u1: SharedTensorBase=None) -> SharedPairBase:
    # f: (SharedTensorBase, SharedTensorBase)->SharedTensorBase
    adjoint_is_None = (x_adjoint is None or y_adjoint is None or u0 is None or u1 is None)
    if adjoint_is_None:
        assert (x_adjoint is None and y_adjoint is None and u0 is None and u1 is None)
        if prf_flag is None:
            prf_flag = StfConfig.prf_flag
        if x.module != y.module:
            raise Exception("must have x.module==y.module")
        if x.owner == y.owner:
            raise Exception("must have x.owner != y.owner")
        with tf.device(StfConfig.RS[0]):
            if prf_flag:
                seedx = get_seed()
                seedy = get_seed()
                seedu = get_seed()
                x_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed=seedx)
                y_adjoint = y.to_SharedTensor_like().random_uniform_adjoint(seed=seedy)
            else:
                x_adjoint = x.to_SharedTensor_like().random_uniform_adjoint()
                y_adjoint = y.to_SharedTensor_like().random_uniform_adjoint()
            u = f(x_adjoint, y_adjoint)
            if not isinstance(u, SharedTensorBase):
                raise StfTypeException("u", "SharedTensorBase", u.tupe)
            if prf_flag:
                u0 = u.random_uniform_adjoint(seed=seedu)
            else:
                u0 = u.random_uniform_adjoint()
            u1 = u - u0
    with tf.device(x.owner):
        if prf_flag and x_adjoint is None:
            x_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed=seedx)
        delta_x = x.to_SharedTensor() - x_adjoint
    with tf.device(y.owner):
        if prf_flag and y_adjoint is None:
            y_adjoint = y.to_SharedTensor_like().random_uniform_adjoint(seed=seedy)
        delta_y = y.to_SharedTensor() - y_adjoint
        v1 = f(delta_x, y.to_SharedTensor()) + u1
    with tf.device(x.owner):
        v0 = f(x_adjoint, delta_y) + u0
    return SharedPairBase(xL=v0, xR=v1, ownerL=x.owner, ownerR=y.owner, fixedpoint=x.fixedpoint + y.fixedpoint)


def BM_PrivateTensor_SharedPair(x: PrivateTensorBase, y: SharedPairBase, f, prf_flag=None,
                                x_adjoint=None, y_adjoint=None, u0=None, u1=None) -> SharedPairBase:
    # f: (PrivateTensorBase, SharedTensorBase)->SharedTensorBase

    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if x.module != y.xL.module:
        raise Exception("must have x.module==y.module")
    if y.ownerL == y.ownerR:
        raise Exception("must have y.ownerL!=y.ownerR")
    if x.owner == y.ownerL:
        adjoint_is_None = (x_adjoint is None or y_adjoint is None or u0 is None or u1 is None)
        assert (x_adjoint is None and y_adjoint is None and u0 is None and u1 is None)
        if adjoint_is_None:
            with tf.device(StfConfig.RS[0]):
                if prf_flag:
                    seedx = get_seed()
                    seedy = get_seed()
                    seedu = get_seed()
                    x_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed=seedx)
                    y_adjoint = y.to_SharedTensor_like().random_uniform_adjoint(seed=seedy)
                else:
                    x_adjoint = x.to_SharedTensor_like().random_uniform_adjoint()
                    y_adjoint = y.xR.random_uniform_adjoint()
                u = f(x_adjoint, y_adjoint)
                if not isinstance(u, SharedTensorBase):
                    raise StfTypeException("u", "SharedTensorBase", u.tupe)
                if prf_flag:
                    u0 = u.random_uniform_adjoint(seed=seedu)
                else:
                    u0 = u.random_uniform_adjoint()
                u1 = u - u0
                # print_op = tf.print("generate random x_adjoint, y_adjoint, u0; compute  u = f(x_adjoint, y_adjoint) "
                #                     "and u1=u-u0")
                # StfConfig.log_op_list.append(print_op)
        with tf.device(x.owner):
            if prf_flag and x_adjoint is None:
                x_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed=seedx)
            delta_x = x.to_SharedTensor() - x_adjoint
            # print_op = tf.print("delta_x = x.to_SharedTensor() - x_adjoint")
            # StfConfig.log_op_list.append(print_op)
        with tf.device(y.ownerR):
            if prf_flag and y_adjoint is None:
                y_adjoint = y.xR.random_uniform_adjoint(seed=seedy)
            delta_y = y.xR - y_adjoint
            v1 = f(delta_x, y.xR)
            v1 = v1 + u1
            # print_op = tf.print("xR = f(delta_x, y.xR) + u1")
            # StfConfig.log_op_list.append(print_op)
        with tf.device(x.owner):
            v0 = f(x_adjoint, delta_y) + u0
            xy0 = f(x.to_SharedTensor(), y.xL)
            # print_op = tf.print("xL = f(x_adjoint, delta_y) + u0 + f(x.to_SharedTensor(), y.xL)")
            # StfConfig.log_op_list.append(print_op)
        xy = SharedPairBase(xL=v0 + xy0, xR=v1, ownerL=y.ownerL, ownerR=y.ownerR,
                            fixedpoint=x.fixedpoint + y.fixedpoint)
    elif x.owner == y.ownerR:
        y = y.mirror()
        xy = BM_PrivateTensor_SharedPair(x, y, f, prf_flag=prf_flag, x_adjoint=x_adjoint, y_adjoint=y_adjoint, u0=u0, u1=u1)
        xy = xy.mirror()
    else:
        raise Exception("must have x.owner==y.ownerL or x.owner==y.ownerR")
    return xy


def BM_SharedPair_PrivateTensor(x: SharedPairBase, y: PrivateTensorBase, f, prf_flag=None,
                                x_adjoint=None, y_adjoint=None, u0=None, u1=None) -> SharedPairBase:
    return BM_PrivateTensor_SharedPair(y, x, lambda a, b: f(b, a), prf_flag, y_adjoint, x_adjoint, u1, u0)


def BM_SharedPair_SharedPair(x: SharedPairBase, y: SharedPairBase, f, prf_flag=None,
                             xL_adjoint=None, xR_adjoin=None, yL_adjoint=None, yR_adjoint=None,
                             uL=None, uR=None) -> SharedPairBase:
    # f: (SharedTensorBase, SharedTensorBase)->SharedTensorBase
    adjoint_is_None = (xL_adjoint is None or xR_adjoin is None or yL_adjoint is None
                       or yR_adjoint is None or uL is None or uR is None)
    if adjoint_is_None:
        assert (xL_adjoint is None and xR_adjoin is None and yL_adjoint is None
                       and yR_adjoint is None and uL is None and uR is None)
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag

    if x.xL.module != y.xL.module:
        raise Exception("must have x.xL.module==y.xL.module")
    if x.ownerL != y.ownerL or x.ownerR != y.ownerR:
        y = y.mirror()
    if x.ownerL != y.ownerL or x.ownerR != y.ownerR:
        raise Exception("must have x.ownerL==y.ownerL and x.ownerR==y.ownerR")
    if adjoint_is_None:
        with tf.device(StfConfig.RS[0]):
            if prf_flag:
                seed_xL = get_seed()
                seed_xR = get_seed()
                seed_yL = get_seed()
                seed_yR = get_seed()
                seed_u = get_seed()
                xL_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed_xL)
                xR_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed_xR)
                yL_adjoint = y.to_SharedTensor_like().random_uniform_adjoint(seed_yL)
                yR_adjoint = y.to_SharedTensor_like().random_uniform_adjoint(seed_yR)
            else:
                xL_adjoint = x.to_SharedTensor_like().random_uniform_adjoint()
                yL_adjoint = y.to_SharedTensor_like().random_uniform_adjoint()
                xR_adjoint = x.to_SharedTensor_like().random_uniform_adjoint()
                yR_adjoint = y.to_SharedTensor_like().random_uniform_adjoint()
            u = f(xL_adjoint + xR_adjoint, yL_adjoint + yR_adjoint)  # SharedTensor
            if not isinstance(u, SharedTensorBase):
                raise StfTypeException("u", "SharedTensorBase", u.tupe)
            if prf_flag:
                uL = u.random_uniform_adjoint(seed_u)
            else:
                uL = u.random_uniform_adjoint()
            uR = u - uL
    with tf.device(x.ownerL):
        if prf_flag and adjoint_is_None:
            xL_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed_xL)
            yL_adjoint = y.to_SharedTensor_like().random_uniform_adjoint(seed_yL)
            uL = u.random_uniform_adjoint(seed_u)
        delta_xL = x.xL - xL_adjoint
        delta_yL = y.xL - yL_adjoint
    with tf.device(y.ownerR):
        if prf_flag and adjoint_is_None:
            xR_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed_xR)
            yR_adjoint = y.to_SharedTensor_like().random_uniform_adjoint(seed_yR)
        delta_xR = x.xR - xR_adjoint
        delta_yR = y.xR - yR_adjoint
    with tf.device(x.ownerL):
        vL = f(delta_xL + delta_xR, delta_yL + delta_yR) + f(xL_adjoint, delta_yL + delta_yR) + \
            f(delta_xL + delta_xR, yL_adjoint) + uL
    with tf.device(x.ownerR):
        vR = f(xR_adjoint, delta_yL + delta_yR) + f(delta_xL + delta_xR, yR_adjoint) + uR
    return SharedPairBase(xL=vL, xR=vR, ownerL=x.ownerL, ownerR=x.ownerR, fixedpoint=x.fixedpoint + y.fixedpoint)



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : compare
   Author : qizhi.zqz
   Email: qizhi.zqz@alibaba-inc.com
   Create Time : 2020-05-14 11:44
   Description : compare : https://eprint.iacr.org/2021/857
"""

from stensorflow.basic.protocol.fnz import fnz_v3 as fnz
from stensorflow.basic.protocol.module_transform import module_transform
from stensorflow.basic.basic_class.base import PrivateTensorBase, SharedPairBase, SharedTensorBase
from stensorflow.global_var import StfConfig
from stensorflow.basic.protocol.ot import assistant_ot
from sympy.ntheory.modular import isprime
import tensorflow as tf


def get_next_prime(n: int):
    while 1:
        if isprime(n):
            return n
        else:
            n += 1


def less_SharedTensorBase_SharedTensorBase(x: SharedTensorBase, y: SharedTensorBase, x_owner,
                                           y_owner) -> SharedPairBase:
    """

    :param x, y:  SharedPair with module=2, shape=[n0, n1, ..., nk]
    :return:  SharedPair with module=2,  shape=[n0, n1, ...nk-1]   x<y in lexicographical order
    """

    if x.module != 2 or y.module != 2:
        raise Exception("must have x.module==2 and y.module==2")
    if x.shape != y.shape:
        raise Exception("must have x.shape==y.shape")
    if x_owner == y_owner:
        raise Exception("must have x_owner!=y_owner")

    p = get_next_prime(x.shape[-1] + 3)

    z = SharedPairBase(ownerL=x_owner, ownerR=y_owner, xL=x, xR=y, fixedpoint=0)
    zp = module_transform(z, p)

    paddings = [[0, 0]] * len(x.shape)
    paddings[-1] = [0, 1]  # add (0, 1) following z[n-1]
    with tf.device(x_owner):
        x_extp = zp.xL.pad(paddings=paddings, constant_values=0)
    with tf.device(y_owner):
        y_ext = y.pad(paddings=paddings)
        y_extp = zp.xR.pad(paddings=paddings, constant_values=1)
    z_extp = SharedPairBase(ownerL=x_owner, ownerR=y_owner, xL=x_extp, xR=y_extp, fixedpoint=0)

    first_nonzero_bit = fnz(z_extp)

    with tf.device(x_owner):
        first_nonzero_bitL = PrivateTensorBase(owner=x_owner, fixedpoint=0,
                                               inner_value=first_nonzero_bit.xL.inner_value,
                                               module=first_nonzero_bit.xL.module)
    with tf.device(y_owner):
        shifted_y_ext = y_ext.lshift(first_nonzero_bit.xR.inner_value)
        p_shifted_y_ext = PrivateTensorBase(owner=y_owner, fixedpoint=0, inner_value=shifted_y_ext.inner_value,
                                            module=shifted_y_ext.module)
    xly = assistant_ot(p_shifted_y_ext, first_nonzero_bitL, prf_flag=StfConfig.prf_flag,
                       compress_flag=StfConfig.compress_flag)
    return xly


def leq_SharedTensorBase_SharedTensorBase(x: SharedTensorBase, y: SharedTensorBase, x_owner, y_owner) -> SharedPairBase:
    """

    :param x, y:  SharedPair with module=2, shape=[n0, n1, ..., nk]
    :return:  SharedPair with module=2,  shape=[n0, n2, ...nk-1]   x<=y in lexicographical order
    """

    if x.module != 2 or y.module != 2:
        raise Exception("must have x.module==2 and y.module==2")
    if x.shape != y.shape:
        raise Exception("must have x.shape==y.shape")
    if x_owner == y_owner:
        raise Exception("must have x_owner!=y_owner")

    p = get_next_prime(x.shape[-1] + 3)  # refer to

    z = SharedPairBase(ownerL=x_owner, ownerR=y_owner, xL=x, xR=y, fixedpoint=0)
    zp = module_transform(z, p)
    paddings = [[0, 0]] * len(x.shape)
    paddings[-1] = [0, 1]  # add (1, 0) following x[n-1]
    with tf.device(x_owner):
        x_extp = zp.xL.pad(paddings=paddings, constant_values=0)
    with tf.device(y_owner):
        y_ext = y.pad(paddings=paddings, constant_values=1)
        y_extp = zp.xR.pad(paddings=paddings, constant_values=1)
    z_extp = SharedPairBase(ownerL=x_owner, ownerR=y_owner, xL=x_extp, xR=y_extp, fixedpoint=0)

    first_nonzero_bit = fnz(z_extp)

    with tf.device(y_owner):

        shifted_y_ext = y_ext.lshift(first_nonzero_bit.xR.inner_value)
        _shifted_y_ext = PrivateTensorBase(owner=y_owner, fixedpoint=0, inner_value=shifted_y_ext.inner_value,
                                           module=shifted_y_ext.module)

    with tf.device(x_owner):

        first_nonzero_bitL = PrivateTensorBase(owner=x_owner, fixedpoint=0,
                                               inner_value=first_nonzero_bit.xL.inner_value,
                                               module=first_nonzero_bit.xL.module)

    xly = assistant_ot(_shifted_y_ext, first_nonzero_bitL, prf_flag=StfConfig.prf_flag,
                       compress_flag=StfConfig.compress_flag)
    return xly


def greater_SharedTensorBase_SharedTensorBase(x: SharedTensorBase, y: SharedTensorBase, x_owner,
                                              y_owner) -> SharedPairBase:
    """

    :param x, y:  SharedPair with module=2, shape=[n0, n1, ..., nk]
    :return:  SharedPair with module=2,  shape=[n0, n2, ...nk-1]   x>y in lexicographical order
    """

    return less_SharedTensorBase_SharedTensorBase(y, x, y_owner, x_owner)


def geq_SharedTensorBase_SharedTensorBase(x: SharedTensorBase, y: SharedTensorBase, x_owner, y_owner) -> SharedPairBase:
    """
    :param x, y:  SharedPair with module=2, shape=[n0, n1, ..., nk]
    :return:  SharedPair with module=2,  shape=[n0, n2, ...nk-1]   x>=y in lexicographical order
    """
    return leq_SharedTensorBase_SharedTensorBase(y, x, y_owner, x_owner)



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : pm1_act
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-06-12 14:43
   Description : (t, x) -> (-1)^tx : https://eprint.iacr.org/2021/857
"""

from stensorflow.basic.basic_class.pair import SharedPairBase
from stensorflow.basic.basic_class.share import SharedTensor, SharedTensorBase
from stensorflow.basic.basic_class.base import PrivateTensorBase
from stensorflow.random.random import get_seed, gen_rint64, gen_rint64_from_seed
from typing import Union
from stensorflow.exception.exception import StfCondException, StfEqualException, StfValueException, StfNoneException
from stensorflow.global_var import StfConfig
import tensorflow as tf


class SharedTensorInt65:
    def __init__(self, x_high: tf.Tensor = None, x_low: tf.Tensor = None, shape=None):
        if x_high is not None and x_low is not None:
            if x_high.dtype != tf.dtypes.int64 or x_low.dtype != tf.dtypes.bool:
                raise StfCondException("x_high.dtype == tf.dtypes.int64 and x_low.dtype != tf.dtypes.bool",
                                       "x_high.dtype ={}, x_low.dtype={}".format(x_high.dtype, x_low.dtype))
            if x_high.shape != x_low.shape:
                raise StfEqualException("x_high.shape", "x_low.shape", x_high.shape, x_low.shape)
            self.x_high = x_high
            self.x_low = x_low
            self.shape = x_high.shape
        elif shape is not None:
            self.shape = shape
        else:
            raise StfNoneException("x_high and x_low or shape")

    @property
    def device(self):
        if self.x_high.device == self.x_low.device:
            return self.x_high.device
        else:
            return [self.x_high.device, self.x_low.device]

    def identity(self):
        x_high = tf.identity(self.x_high)
        x_low = tf.identity(self.x_low)
        return SharedTensorInt65(x_high=x_high, x_low=x_low)

    @classmethod
    def load_from_SharedTensor(cls, x: SharedTensorBase):
        x_low = tf.cast(x.inner_value % 2, 'bool')
        x_high = x.inner_value // 2
        return SharedTensorInt65(x_high=x_high, x_low=x_low)

    @classmethod
    def load_from_PrivateTensor(cls, x: PrivateTensorBase):
        x_low = tf.cast(x.inner_value % 2, 'bool')
        x_high = x.inner_value // 2
        return SharedTensorInt65(x_high=x_high, x_low=x_low)

    def to_SharedTensor(self):
        return SharedTensor(inner_value=self.x_high * 2 + tf.cast(self.x_low, 'int64'))

    def zeros_like(self):
        x_low = tf.logical_and(self.x_low, False)
        x_high = tf.zeros_like(self.x_high)
        return SharedTensorInt65(x_high=x_high, x_low=x_low)

    def __add__(self, other):
        x_low = tf.logical_or(tf.logical_and(self.x_low, tf.logical_not(other.x_low)),
                              tf.logical_and(tf.logical_not(self.x_low), other.x_low))
        carry = tf.logical_and(self.x_low, other.x_low)
        x_high = self.x_high + other.x_high + tf.cast(carry, 'int64')
        return SharedTensorInt65(x_high=x_high, x_low=x_low)

    def __neg__(self):
        x_low = self.x_low
        x_high = -self.x_high - tf.cast(x_low, 'int64')
        return SharedTensorInt65(x_high=x_high, x_low=x_low)

    def __sub__(self, other):
        return self + (-other)

    def random_uniform_adjoint(self, seed=None):
        """

        :param seed: seed for generate random number
        :return:     generate  pseudo-random SharedTensorInt65 with same shape and
                    module with self
        """
        if seed is not None:
            x = gen_rint64_from_seed(shape=[2] + self.shape, seed=seed)
        else:
            x = gen_rint64([2] + self.shape)
        x_high = x[0, ...]
        x_low = tf.cast(x[1, ...] % 2, 'bool')
        adjoint = SharedTensorInt65(x_high=x_high, x_low=x_low)
        return adjoint


class SharedPairInt65:
    def __init__(self, ownerL, ownerR, xL: SharedTensorInt65 = None, xR: SharedTensorInt65 = None,
                 fixedpoint=StfConfig.default_fixed_point):
        self.ownerL = ownerL
        self.ownerR = ownerR
        if xL is not None and xR is not None:
            with tf.device(ownerL):
                if xL.device == ownerL.to_string():
                    self.xL = xL
                else:
                    self.xL = xL.identity()
                self.shape = self.xL.shape
            with tf.device(ownerR):
                if xR.device == ownerR.to_string():
                    self.xR = xR
                else:
                    self.xR = xR.identity()
                self.ownerR = ownerR

        else:
            self.xL = None
            self.xR = None
            self.shape = None
        self.fixedpoint = fixedpoint

    @classmethod
    def from_SharedPair(cls, x: SharedPairBase):
        ownerL = x.ownerL
        ownerR = x.ownerR
        with tf.device(ownerL):
            xL = SharedTensorInt65.load_from_SharedTensor(x.xL)
        with tf.device(ownerR):
            xR = SharedTensorInt65.load_from_SharedTensor(x.xR)
        return SharedPairInt65(ownerL, ownerR, xL, xR, x.fixedpoint)

    @classmethod
    def load_from_SharedTensorInt65(cls, x: SharedTensorInt65, owner, other_owner, fixedpoint=0):
        with tf.device(owner):
            xL = x.identity()
        with tf.device(other_owner):
            xR = x.zeros_like()
        return SharedPairInt65(owner, other_owner, xL, xR, fixedpoint)

    def to_SharedPair(self):
        with tf.device(self.ownerL):
            xL = self.xL.to_SharedTensor()
        with tf.device(self.ownerR):
            xR = self.xR.to_SharedTensor()
        return SharedPairBase(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=self.fixedpoint)

    def half_of_even(self) -> SharedPairBase:
        with tf.device(self.ownerL):
            xL = SharedTensorBase(inner_value=self.xL.x_high)
        with tf.device(self.ownerR):
            xR = SharedTensorBase(inner_value=self.xR.x_high + tf.cast(self.xR.x_low, 'int64'))
        return SharedPairBase(ownerL=self.ownerL, ownerR=self.ownerR,
                              xL=xL, xR=xR, fixedpoint=self.fixedpoint)

    def to_tf_tensor(self, owner=None):
        z = self.to_SharedPair()
        return z.to_tf_tensor(owner)

    def mirror(self):
        ownerR = self.ownerL
        ownerL = self.ownerR
        xR = self.xL
        xL = self.xR
        return SharedPairInt65(ownerL, ownerR, xL, xR, self.fixedpoint)

    def __add__(self, other):
        if self.fixedpoint != other.fixedpoint:
            raise StfEqualException("self.fixedpoint", "other.fixedpoint", self.fixedpoint, other.fixedpoint)
        if self.ownerL == other.ownerR and self.ownerR == other.ownerL:
            other = other.mirror()
        with tf.device(self.ownerL):
            xL = self.xL + other.xL
        with tf.device(self.ownerR):
            xR = self.xR + other.xR
        return SharedPairInt65(ownerL=self.ownerL, ownerR=self.ownerR, xL=xL, xR=xR, fixedpoint=self.fixedpoint)

    def __neg__(self):
        return SharedPairInt65(ownerL=self.ownerL, ownerR=self.ownerR, xL=-self.xL, xR=-self.xR,
                               fixedpoint=self.fixedpoint)

    def __sub__(self, other):
        return self + (-other)

    def to_SharedTensor_like(self):
        return SharedTensorInt65(shape=self.shape)


def _pm1_act(t: SharedTensor, x: Union[SharedTensor, SharedTensorInt65]) -> Union[SharedTensor, SharedTensorInt65]:
    # (t, x) -> (-1)^tx
    if t.module != 2:
        raise StfValueException("t.module", 2, t.module)
    if t.shape != x.shape:
        raise StfEqualException("t.shape", "x.shape", t.shape, x.shape)
    if isinstance(x, SharedTensorBase):
        inner_value = (x.inner_value - t.inner_value * 2 * x.inner_value)
        return SharedTensor(inner_value=inner_value, module=x.module)
    elif isinstance(x, SharedTensorInt65):
        x_low = x.x_low
        x_high = (1 - 2 * t.inner_value) * x.x_high - t.inner_value * tf.cast(x.x_low, 'int64')
        return SharedTensorInt65(x_high=x_high, x_low=x_low)


def pm1_act(t: SharedTensor, x: Union[SharedTensor, SharedTensorInt65], t_owner, x_owner, RS_owner,
            prf_flag=None, compress_flag=None) \
        -> Union[SharedPairBase, SharedPairInt65]:
    # Step 1.  generate t_adjoint, x_adjoint, b_t, b_x, s.t  (-1)^t_adjoint x_adjoint = b_t + b_x
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag
    with tf.device(RS_owner):
        if prf_flag:
            seed_t = get_seed()
            seed_x = get_seed()
            seed_b = get_seed()
        else:
            seed_t = None
            seed_x = None
            seed_b = None
        t_adjoint = t.random_uniform_adjoint(seed=seed_t)
        x_adjoint = x.random_uniform_adjoint(seed=seed_x)
        b = _pm1_act(t_adjoint, x_adjoint)
        b_t = b.random_uniform_adjoint(seed=seed_b)
        b_x = b - b_t
        if compress_flag and not prf_flag:
            t_adjoint_compress = t_adjoint.to_compress_tensor()

    # Step2. compute delta_t:=t-t_adjoint
    with tf.device(t_owner):
        if prf_flag:
            t_adjoint = t.random_uniform_adjoint(seed=seed_t)
            b_t = b.random_uniform_adjoint(seed=seed_b)
        elif compress_flag:
            t_adjoint.decompress_from(t_adjoint_compress)

        delta_t = t - t_adjoint
        if compress_flag:
            delta_t_compress = delta_t.to_compress_tensor()

    # Step 3. compute delta_x:=x-x_adjoint
    with tf.device(x_owner):
        if prf_flag:
            x_adjoint = x.random_uniform_adjoint(seed=seed_x)
        delta_x = x - x_adjoint

    # Step4. compute  y_t:=(-1)^t(x-x_adjoint)+(-1)^delta_t b_t
    with tf.device(t_owner):
        y_t = _pm1_act(t, delta_x) + _pm1_act(delta_t, b_t)

    # Step 5. compute y_x:=(-1)^delta_t + b_x
    with tf.device(x_owner):
        if compress_flag:
            delta_t.decompress_from(delta_t_compress)
        y_x = _pm1_act(delta_t, b_x)
    if isinstance(x, SharedTensor):
        return SharedPairBase(ownerL=t_owner, ownerR=x_owner, xL=y_t, xR=y_x, fixedpoint=0)
    elif isinstance(x, SharedTensorInt65):
        return SharedPairInt65(ownerL=t_owner, ownerR=x_owner, xL=y_t, xR=y_x, fixedpoint=0)


def pm1_pair_act(t: SharedPairBase, x: Union[SharedTensor, SharedTensorInt65], x_owner, RS_owner,
                 prf_flag=None, compress_flag=None) \
        -> Union[SharedPairBase, SharedPairInt65]:
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag
    if t.ownerL == x_owner and t.ownerR != x_owner:
        t = t.mirror()
    if t.ownerR == x_owner:
        pass
    else:
        raise StfCondException("t.ownerL==x_owner or t.ownerR==x.owner",
                               "t.ownerL={}, t.ownerR={}, x_owner={}".format(t.ownerL, t.ownerR, x_owner))
    with tf.device(x_owner):
        tR_act_x = _pm1_act(t.xR, x)
    return pm1_act(t.xL, tR_act_x, t_owner=t.ownerL, x_owner=x_owner, RS_owner=RS_owner,
                   prf_flag=prf_flag, compress_flag=compress_flag)


def pm1_pair_act_pair(t: SharedPairBase, x: Union[SharedPairBase, SharedPairInt65], RS_owner, prf_flag=None,
                      compress_flag=None) -> Union[SharedPairBase, SharedPairInt65]:
    """

    :param compress_flag:
    :param t:
    :param x:
    :param RS_owner:
    :return: (-1)^tx
    """
    if prf_flag is None:
        prf_flag = StfConfig.prf_flag
    if compress_flag is None:
        compress_flag = StfConfig.compress_flag

    if t.ownerL != x.ownerL or t.ownerR != x.ownerR:
        if t.ownerL != x.ownerR or t.ownerR != x.ownerL:
            raise Exception(
                "must have t.ownerL==x.ownerL and t.ownerR==x.ownerR or t.ownerL==x.ownerR and t.ownerR==x.ownerL")
        else:
            x = x.mirror()
    with tf.device(RS_owner):
        if prf_flag:
            seed_tL = get_seed()
            seed_tR = get_seed()
            seed_xL = get_seed()
            seed_xR = get_seed()
            seed_y = get_seed()
            tL_adjoint = t.to_SharedTensor_like().random_uniform_adjoint(seed_tL)
            tR_adjoint = t.to_SharedTensor_like().random_uniform_adjoint(seed_tR)
            xL_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed_xL)
            xR_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed_xR)
        else:
            tL_adjoint = t.to_SharedTensor_like().random_uniform_adjoint()
            tR_adjoint = t.to_SharedTensor_like().random_uniform_adjoint()
            xL_adjoint = x.to_SharedTensor_like().random_uniform_adjoint()
            xR_adjoint = x.to_SharedTensor_like().random_uniform_adjoint()
        y = _pm1_act(tL_adjoint + tR_adjoint, xL_adjoint + xR_adjoint)
        if prf_flag:
            yL = y.random_uniform_adjoint(seed=seed_y)
        else:
            yL = y.random_uniform_adjoint()
        yR = y - yL
        if compress_flag:
            if not prf_flag:
                tL_adjoint_compressed = tL_adjoint.to_compress_tensor()
                tR_adjoint_compressed = tR_adjoint.to_compress_tensor()
    with tf.device(t.ownerL):
        if prf_flag:
            tL_adjoint = t.to_SharedTensor_like().random_uniform_adjoint(seed_tL)
            xL_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed_xL)
        elif compress_flag:
            tL_adjoint.decompress_from(tL_adjoint_compressed)
        delta_tL = t.xL - tL_adjoint
        delta_xL = x.xL - xL_adjoint
        zL = _pm1_act(t.xL, delta_xL)
        if compress_flag:
            delta_tL_compressed = delta_tL.to_compress_tensor()
    with tf.device(t.ownerR):
        if prf_flag:
            tR_adjoint = t.to_SharedTensor_like().random_uniform_adjoint(seed_tR)
            xR_adjoint = x.to_SharedTensor_like().random_uniform_adjoint(seed_xR)
        elif compress_flag:
            tR_adjoint.decompress_from(tR_adjoint_compressed)
        delta_tR = t.xR - tR_adjoint
        delta_xR = x.xR - xR_adjoint
        zR = _pm1_act(t.xR, delta_xR)
        if compress_flag:
            delta_tR_compressed = delta_tR.to_compress_tensor()
    with tf.device(t.ownerL):
        if compress_flag:
            delta_tR_new = delta_tR.decompress_from_to_new(delta_tR_compressed)
            uL = _pm1_act(t.xL, zR) + _pm1_act(delta_tL + delta_tR_new, yL)
        else:
            uL = _pm1_act(t.xL, zR) + _pm1_act(delta_tL + delta_tR, yL)
    with tf.device(t.ownerR):
        if compress_flag:
            delta_tL_new = delta_tL.decompress_from_to_new(delta_tL_compressed)
            uR = _pm1_act(t.xR, zL) + _pm1_act(delta_tL_new + delta_tR, yR)
        else:
            uR = _pm1_act(t.xR, zL) + _pm1_act(delta_tL + delta_tR, yR)
    if isinstance(x, SharedPairBase):
        return SharedPairBase(ownerL=t.ownerL, ownerR=t.ownerR, xL=uL, xR=uR, fixedpoint=x.fixedpoint)
    elif isinstance(x, SharedPairInt65):
        return SharedPairInt65(ownerL=t.ownerL, ownerR=t.ownerR, xL=uL, xR=uR, fixedpoint=x.fixedpoint)



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : Input
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-09-11 19:43
   Description : description what the main function of this file
"""

from stensorflow.ml.nn.layers.layer import Layer
# import numpy as np
from stensorflow.basic.basic_class.private import PrivateTensor
from stensorflow.basic.basic_class.pair import SharedPair
from typing import Union, List


# from stensorflow.global_var import workerL, workerR

class Input(Layer):
    """
    Input Layer
    """
    def __init__(self, dim: Union[int, List[int]], x: PrivateTensor, owner=None):
        if isinstance(dim, int):
            if x.shape[1] != dim:
                raise Exception("must have x.shape[1] == dim")
        else:
            if x.shape[1:] != dim:
                raise Exception("must have x.shape[1:] == dim, but x.shape[1:]={}".format(x.shape[1:]))
        super(Input, self).__init__(output_dim=dim, fathers=[])
        self.y = x
        # self.ploss_px = {None: x.zeros_like().transpose()}
        if owner is None:
            self.owner = x.owner
        else:
            self.owner = owner

    def forward(self):
        return

    def backward(self):
        # if len(self.ploss_px) > 0:
        #     return
        for child in self.children:
            if isinstance(child, Layer):
                child.backward()
            else:
                raise Exception("children must be a Layer")
        # self.ploss_px = {None: x.zeros_like().transpose()}

    def replace(self, x: Union[SharedPair, PrivateTensor]):
        if x.shape[1] != self.output_dim:
            raise Exception("must have x.shape[1] == self.output_dim")
        self.y = x



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2021 All Rights Reserved.
   ------------------------------------------------------
   File Name : homo_mat_mul
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2021/9/30 下午5:31
   Description : description what the main function of this file
"""


from stensorflow.basic.basic_class.base import PrivateTensorBase, SharedPairBase, SharedTensorBase
from stensorflow.exception.exception import StfEqualException, StfCondException
import tensorflow as tf
from typing import Union
from stensorflow.basic.protocol.bilinear_map import BM_PrivateTensor_SharedPair, BM_PrivateTensor_PrivateTensor, \
    BM_SharedPair_SharedPair, BM_SharedPair_PrivateTensor
from stensorflow.homo_enc.homo_enc import gene_key, enc, vec_mul_vec_to_share, dec



def vecmulvec_privat_private(x: PrivateTensorBase, y: PrivateTensorBase, fixed_point=None) -> SharedPairBase:
    if isinstance(x, PrivateTensorBase) and isinstance(y, PrivateTensorBase):
        pass
    else:
        raise NotImplementedError
    if x.shape != y.shape:
        raise StfEqualException("x.shape", "y.shape", x.shape, y.shape)
    if x.owner == y.owner:
        return PrivateTensorBase.__mul__(x, y, fixed_point)
    else:
        pass
    vec_len = x.shape[0]
    with tf.device(x.owner):
        sk, pk, gk = gene_key()
        cipher_vec_x = enc(pk, x.inner_value)
    with tf.device(y.owner):
        cipher_out_vec, share_out_vec = vec_mul_vec_to_share(pk, y.inner_value, cipher_vec_x)

        xR = SharedTensorBase(inner_value=share_out_vec)
    with tf.device(x.owner):
        plaintext_out_vec = dec(sk, [vec_len], cipher_out_vec)
        xL = SharedTensorBase(inner_value=plaintext_out_vec)

    r = SharedPairBase(ownerL=x.owner, ownerR=y.owner, xL=xL, xR=xR,
                       fixedpoint=x.fixedpoint + y.fixedpoint)
    if fixed_point is None:
        return r
    else:
        return r.dup_with_precision(new_fixedpoint=fixed_point)


def mul_homo_privat_private(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase], fixed_point=None) -> SharedPairBase:
    if len(x.shape) == 1 and len(y.shape) == 1:
        return vecmulvec_privat_private(x, y, fixed_point)
    else:
        reshaped_x = x.reshape([-1])
        reshaped_y = y.reshape([-1])
        z = vecmulvec_privat_private(reshaped_x, reshaped_y, fixed_point=fixed_point)
        return z.reshape(x.shape)



def mul_homo(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase], fixed_point=None) -> SharedPairBase:
    y_is_private = isinstance(y, PrivateTensorBase)
    x_is_private = isinstance(x, PrivateTensorBase)

    if y_is_private and x_is_private:
        return mul_homo_privat_private(x, y, fixed_point)
    else:
        raise NotImplementedError

def mul_homo_offline(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase]):
    print("matmul_homo_offline")
    x.serialize()
    y.serialize()
    z = mul_homo(x, y)
    z.serialize()
    return z


def mul_homo_online(x: Union[PrivateTensorBase, SharedPairBase], y: Union[PrivateTensorBase, SharedPairBase],
                       x_adjoint: Union[PrivateTensorBase, SharedPairBase], y_adjoint: Union[PrivateTensorBase, SharedPairBase],  z_adjoint: Union[PrivateTensorBase, SharedPairBase], fixed_point=None):
    x_is_private = isinstance(x, PrivateTensorBase)
    x_is_shared = isinstance(x, SharedPairBase)
    y_is_private = isinstance(y, PrivateTensorBase)
    y_is_shared = isinstance(y, SharedPairBase)

    if x_is_private and y_is_private:
        assert isinstance(x_adjoint, PrivateTensorBase) and isinstance(y_adjoint, PrivateTensorBase)
        z = BM_PrivateTensor_PrivateTensor(x, y, lambda _x, _y: SharedTensorBase.__mul__(_x, _y),
                                           prf_flag=False, x_adjoint=x_adjoint.to_SharedTensor(), y_adjoint=y_adjoint.to_SharedTensor(),
                                           u0=z_adjoint.xL, u1=z_adjoint.xR)
    elif x_is_private and y_is_shared:
        assert isinstance(x_adjoint, PrivateTensorBase) and isinstance(y_adjoint, PrivateTensorBase)
        z = BM_PrivateTensor_SharedPair(x, y, lambda _x, _y: SharedTensorBase.__mul__(_x, _y),
                                        prf_flag=False, x_adjoint=x_adjoint.to_SharedTensor(),
                                        y_adjoint=y_adjoint.to_SharedTensor(), u0=z_adjoint.xL, u1=z_adjoint.xR
                                        )
    elif x_is_shared and y_is_private:
        assert isinstance(x_adjoint, PrivateTensorBase) and isinstance(y_adjoint, PrivateTensorBase)
        z = BM_SharedPair_PrivateTensor(x, y, lambda _x, _y: SharedTensorBase.__mul__(_x, _y),
                                        prf_flag=False, x_adjoint=x_adjoint.to_SharedTensor(),
                                        y_adjoint=y_adjoint.to_SharedTensor(),
                                        u0=z_adjoint.xL, u1=z_adjoint.xR)
    elif x_is_shared and y_is_shared:
        assert isinstance(x_adjoint, SharedPairBase) and isinstance(y_adjoint, SharedPairBase)
        z = BM_SharedPair_SharedPair(x, y, lambda _x, _y: SharedTensorBase.__mul__(_x, _y),
                                     prf_flag=False, xL_adjoint=x_adjoint.xL, xR_adjoin=x_adjoint.xR,
                                     yL_adjoint=y_adjoint.xL, yR_adjoint=y_adjoint.xR,
                                     uL=z_adjoint.xL, uR=z_adjoint.xR)
    else:
        raise StfCondException("x, y are PrivateTensor or SharedPair", "x is, y is".format(type(x), type(y)))
    return z






#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group
   Copyright (c) 2004-2020 All Rights Reserved.
   ------------------------------------------------------
   File Name : Layer
   Author : Qizhi Zhang
   Email: qizhi.zqz@antgroup.com
   Create Time : 2020-09-11 11:51
   Description : description what the main function of this file
"""

from functools import reduce


class Layer:
    def __init__(self, output_dim, fathers=None):
        self.output_dim = output_dim
        self.fathers = fathers
        self.children = []
        self.w = []  # list
        self.x = []  # list
        self.y = None
        self.ploss_pw = []  # list
        self.ploss_px = {}  # map Layer-> ploss_py * py_px
        self.ploss_py = None

    def func(self, w, x):
        # (w, x)->y
        pass

    def pull_back(self, w, x, y, ploss_py):
        # (w, x, y, ploss_py)-> (ploss_pw, ploss_px)
        pass

    def __hash__(self):
        return sum(hash(q) for q in self.w)

    def __eq__(self, other):
        return self.w == other.w

    def add_child(self, child):
        if child not in self.children:
            self.children.append(child)

    def forward(self):
        if self.y is not None:
            return
        for father in self.fathers:
            if isinstance(father, Layer):
                # try:
                father.forward()
                # except Exception as e:

            else:
                raise Exception("father must be a layer")
        self.x = list(map(lambda father: father.y, self.fathers))
        self.y = self.func(self.w, self.x)

    def backward(self):
        if len(self.ploss_px) > 0:
            return
        for child in self.children:
            if isinstance(child, Layer):
                child.backward()
            else:
                raise Exception("children must be a Layer")

        if self.ploss_py is None:
            ploss_px_of_children = list(map(lambda childre: child.ploss_px[self], self.children))

            self.ploss_py = reduce(lambda x, y: x + y, ploss_px_of_children)

        self.ploss_pw, self.ploss_px = self.pull_back(self.w, self.x, self.y, self.ploss_py)

    def cut_off(self):
        self.y = None



#!/usr/bin/env python
# coding=utf-8
"""
   Ant Group Copyright (c) 2021
   Copyright 2016 The TensorFlow Authors
   All Rights Reserved.
"""
from stensorflow.ml.nn.layers.layer import Layer
from typing import Union, List
from stensorflow.basic.basic_class.private import PrivateTensor
from stensorflow.basic.basic_class.pair import SharedVariablePair, SharedPair
from tensorflow.python.keras.utils import conv_utils
from stensorflow.basic.operator.poolingop import avg_pool2d, sum_pool2d_grad, max_pool2d, max_pool2d_back
import tensorflow as tf
from stensorflow.global_var import StfConfig




PModel = StfConfig.pool_module

class AveragePooling2D(Layer):
    """
    Average pooling operation for spatial data.
     Argument:
        output_dim:
        fathers:
        pool_size:
            An integer or tuple/list of 2 integers: (pool_height, pool_width) specifying the size of the pooling window.
            Can be a single integer to specify the same value for all spatial dimensions.
        strides:
            An integer or tuple/list of 2 integers, specifying the strides of the pooling operation.
            Can be a single integer to specify the same value for all spatial dimensions.
        padding:
            A string. The padding method, only support `"VALID"`
        data_format:
            default format "NHWC", the data is stored in the order of: [batch, height, width, channels].
    """
    def __init__(self, output_dim,
                 fathers: List[Layer],
                 pool_size=(2, 2),
                 strides=None,
                 padding='VALID',
                 data_format=None):
        if len(fathers) != 1:
            raise Exception("must have len(fathers) == 1 ")
        if data_format is None:
            data_format = "NHWC"
        if strides is None:
            strides = pool_size
        # 初始化参数
        # 池化尺寸
        self.pool_size = conv_utils.normalize_tuple(pool_size, 2, 'pool_size')
        self.strides = conv_utils.normalize_tuple(strides, 2, "strides")
        self.padding = padding
        self.data_format = data_format
        # 保存传入数据的shape
        self.input_shape = fathers[0].output_dim
        if not output_dim:
            output_dim = self.compute_shape()
        super(AveragePooling2D, self).__init__(output_dim=output_dim, fathers=fathers)

    def func(self, w: List[SharedVariablePair], x: List[Union[PrivateTensor, SharedPair]]):
        """
        Forward propagation
        :param w: None
        :param x: inputs， PrivateTensor or SharedPair
        :return: Has the same type of inputs
        """
        if len(x) != 1:
            raise Exception("must have len(x) == 1")
        if self.data_format == 'NHWC':
            pool_shape = (1,) + self.pool_size + (1,)
            strides = (1,) + self.strides + (1,)
        else:
            pool_shape = (1, 1) + self.pool_size
            strides = (1, 1) + self.strides
        # 更新self.input_shape
        self.input_shape = x[0].shape
        return avg_pool2d(input=x[0], ksize=pool_shape, strides=strides, padding=self.padding,
                          data_format=self.data_format)

    def pull_back(self,
                  w: List[SharedPair],
                  x: List[Union[PrivateTensor, SharedPair]],
                  y: SharedPair,
                  ploss_py: SharedPair):
        """
        Back propagation
        :param w: None
        :param x: inputs
        :param y:
        :param ploss_py:
        :return:
        """
        if len(x) != 1:
            raise Exception("must have len(x)==1")
        if self.data_format == 'NHWC':
            pool_shape = (1,) + self.pool_size + (1,)
            strides = (1,) + self.strides + (1,)
        else:
            pool_shape = (1, 1) + self.pool_size
            strides = (1, 1) + self.strides
        # 调用反向传播
        # 计算反向传播系数
        N = self.pool_size[0] * self.pool_size[1]
        # 更新ploss_py
        ploss_py = ploss_py / N
        res = sum_pool2d_grad(input_shape=self.input_shape, out_backprop=ploss_py,
                              ksize=pool_shape, strides=strides, padding=self.padding)
        ploss_px = {self.fathers[0]: res}
        return [], ploss_px

    def compute_shape(self):
        if self.data_format == "NHWC":
            rows = self.input_shape[0]
            cols = self.input_shape[1]
        else:
            rows = self.input_shape[1]
            cols = self.input_shape[2]
        rows = conv_utils.conv_output_length(rows, self.pool_size[0], self.padding.lower(),
                                             self.strides[0])
        cols = conv_utils.conv_output_length(cols, self.pool_size[1], self.padding.lower(),
                                             self.strides[1])
        if self.data_format == "NHWC":
            return [rows, cols, self.input_shape[2]]
        else:
            return [self.input_shape[0], rows, cols]


class AveragePooling2DLocal(Layer):
    def __init__(self, output_dim,
                 fathers: List[Layer],
                 owner,
                 pool_size=(2, 2),
                 strides=None,
                 padding='VALID',
                 data_format=None):
        if len(fathers) != 1:
            raise Exception("must have len(fathers) == 1 ")

        if data_format is None:
            data_format = "NHWC"
        if strides is None:
            strides = pool_size
        # 初始化参数
        self.owner = owner
        # 池化尺寸
        self.pool_size = conv_utils.normalize_tuple(pool_size, 2, 'pool_size')
        self.strides = conv_utils.normalize_tuple(strides, 2, "strides")
        self.padding = padding
        self.data_format = data_format
        # 保存传入数据的shape
        self.input_shape = None
        self.input_shape = fathers[0].output_dim
        if not output_dim:
            output_dim = self.compute_shape()
        super(AveragePooling2DLocal, self).__init__(output_dim=output_dim, fathers=fathers)

    def func(self, w: List[PrivateTensor], x: [PrivateTensor]):
        """
        Forward propagation
        :param w: None
        :param x: inputs
        :return: Has the same type of inputs
        """
        if len(x) != 1:
            raise Exception("must have len(x) == 1")
        if self.data_format == 'NHWC':
            pool_shape = (1,) + self.pool_size + (1,)
            strides = (1,) + self.strides + (1,)
        else:
            pool_shape = (1, 1) + self.pool_size
            strides = (1, 1) + self.strides
        x = x[0].to_private(owner=self.owner)  # x为PrivaterTensor
        # 更新self.input_shape
        self.input_shape = x.shape
        # return PrivateTensor
        res = avg_pool2d(input=x, ksize=pool_shape, strides=strides, padding=self.padding,
                         data_format=self.data_format)
        # 下面的代码是需要进行逐层测试时，将需要向下一层传播的信息转化为ShairPair，来符合整体框架使用的
        # ans = SharedPair(ownerL="L", ownerR="R", shape=res.shape)
        # ans.load_from_tf_tensor(res.to_tf_tensor())
        # return ans
        return res

    def pull_back(self,
                  w: List[PrivateTensor],
                  x: [PrivateTensor],
                  y: SharedPair,
                  ploss_py: SharedPair):
        """
        Back propagation
        :param w: None
        :param x: inputs
        :param y:
        :param ploss_py:
        :return:
        """
        if len(x) != 1:
            raise Exception("must have len(x)==1")
        if self.data_format == 'NHWC':
            pool_shape = (1,) + self.pool_size + (1,)
            strides = (1,) + self.strides + (1,)
        else:
            pool_shape = (1, 1) + self.pool_size
            strides = (1, 1) + self.strides
        # 调用反向传播
        # 计算反向传播系数
        N = self.pool_size[0] * self.pool_size[1]
        fix = ploss_py.fixedpoint
        # 将数据转化为tensor
        ploss_py = ploss_py.to_tf_tensor(owner=self.owner)
        ploss_py = ploss_py/N
        # 将数据扩大
        ploss_py = ploss_py*(2**fix)
        ploss_py = tf.cast(ploss_py, dtype=tf.int64)
        # compute
        with tf.device(self.owner):
            res = PModel.sum_pool_grad(self.input_shape, grad=ploss_py,
                                       ksize=pool_shape, strides=strides, padding=self.padding)
        # 将数据缩小
        res = res/(2**fix)
        zx = PrivateTensor(owner=self.owner)
        zx.load_from_tf_tensor(res)
        # 下面的代码是需要进行逐层测试时，将需要向下一层传播的信息转化为ShairPair，来符合整体框架使用的
        # zx = SharedPair(ownerL="L", ownerR="R", shape=res.shape)
        # zx.load_from_tf_tensor(res)
        # 将数据转化为private
        ploss_px = {self.fathers[0]: zx}
        return [], ploss_px

    def compute_shape(self):
        # 假设采用的是更新后的input_shape
        if self.data_format == "NHWC":
            rows = self.input_shape[0]
            cols = self.input_shape[1]
        else:
            rows = self.input_shape[1]
            cols = self.input_shape[2]
        rows = conv_utils.conv_output_length(rows, self.pool_size[0], self.padding.lower(),
                                             self.strides[0])
        cols = conv_utils.conv_output_length(cols, self.pool_size[1], self.padding.lower(),
                                             self.strides[1])
        if self.data_format == "NHWC":
            return [rows, cols, self.input_shape[2]]
        else:
            return [self.input_shape[0], rows, cols]



class MaxPooling2D(Layer):
    """
    Average pooling operation for spatial data.
     Argument:
        output_dim:
        fathers:
        pool_size:
            An integer or tuple/list of 2 integers: (pool_height, pool_width) specifying the size of the pooling window.
            Can be a single integer to specify the same value for all spatial dimensions.
        strides:
            An integer or tuple/list of 2 integers, specifying the strides of the pooling operation.
            Can be a single integer to specify the same value for all spatial dimensions.
        padding:
            A string. The padding method, only support `"VALID"`
        data_format:
            default format "NHWC", the data is stored in the order of: [batch, height, width, channels].
    """
    def __init__(self, output_dim,
                 fathers: List[Layer],
                 pool_size=(2, 2)):
        if len(fathers) != 1:
            raise Exception("must have len(fathers) == 1 ")
        self.ksize = [1, pool_size[0], pool_size[1], 1]
        # 保存传入数据的shape
        self.input_shape = fathers[0].output_dim
        if not output_dim:
            output_dim = self.compute_shape()
        super(MaxPooling2D, self).__init__(output_dim=output_dim, fathers=fathers)
        self.pool_size = pool_size
        self.index_list = None

    def func(self, w: List[SharedVariablePair], x: List[SharedPair]):
        """
        Forward propagation
        :param w: None
        :param x: inputs， PrivateTensor or SharedPair
        :return: Has the same type of inputs
        """
        if len(x) != 1:
            raise Exception("must have len(x) == 1")
        y, self.index_list = max_pool2d(x[0], ksize=self.ksize, return_s=True)
        return y

    def pull_back(self,
                  w: List[SharedPair],
                  x: List[Union[PrivateTensor, SharedPair]],
                  y: SharedPair,
                  ploss_py: SharedPair):
        """
        Back propagation
        :param w: None
        :param x: inputs
        :param y:
        :param ploss_py:
        :return:
        """
        if len(x) != 1:
            raise Exception("must have len(x)==1")
        res = max_pool2d_back(ploss_py=ploss_py, ksize=self.ksize, index_list=self.index_list)
        ploss_px = {self.fathers[0]: res}
        return [], ploss_px

    def compute_shape(self):
        return [self.input_shape[0]//self.ksize[1], self.input_shape[1]//self.ksize[2], self.input_shape[2]]












import setuptools
import subprocess
from setuptools.command.install import install
import os

def _copy_cops():
    print("_compile_cops:", os.getcwd())
    home = os.environ['HOME']
    os.makedirs(home + '/stf_cops', exist_ok=True)
    subprocess.check_call('cp -r ./cops/* ~/stf_cops/', shell=True)
    cops_path = home + '/stf_cops/'
    print("cops_path=", cops_path)
    #subprocess.check_call('sh compile.sh', cwd=cops_path, shell=True)


class MorseSTFInstall(install):
    def run(self):
        super().run()
        _copy_cops()


# TODO: we need a license here
setuptools.setup(
    name="morse-stf",
    version="0.1.34",
    author="Morse-STF Team",
    author_email="morse-stf@service.alipay.com",
    description="Morse Secure TensorFlow",
    url="https://github.com/alipay/Antchain-MPC/morse-stf",
    install_requires=[
        'matplotlib==3.3.2',
        'tensorflow==2.2.0',
        'pandas==1.0.5',
        'sympy==1.6',
        'scikit-learn==0.23.1'
    ],


    entry_points={
        'console_scripts': [
            'morse-stf-server=stensorflow.engine.start_server:main',
        ],
    },
    cmdclass={
        'install': MorseSTFInstall,
    },
    classifiers=[
        "Programming Language :: Python :: 3",
    ],
    packages=setuptools.find_packages(include=['stensorflow', 'stensorflow.*']),
    python_requires=">=3.6",
)


#
# install_requires=[
#         'absl-py==0.9.0',
#         'astor==0.8.1',
#         'astunparse==1.6.3',
#         'attrs==19.3.0',
#         'cachetools==4.1.0',
#         'certifi==2020.12.5',
#         'chardet==3.0.4',
#         'click==7.1.2',
#         'coverage==5.1',
#         'cycler==0.10.0',
#         'Cython==0.29.21',
#         'dataclasses==0.8',
#         'dill==0.3.3',
#         'Flask==1.1.2',
#         'future==0.18.2',
#         'gast==0.3.3',
#         'google-auth==1.13.1',
#         'google-auth-oauthlib==0.4.1',
#         'google-pasta==0.2.0',
#         'googleapis-common-protos==1.53.0',
#         'grpcio==1.28.1',
#         'h5py==2.10.0',
#         'idna==2.9',
#         'importlib-metadata==1.7.0',
#         'importlib-resources==5.1.2',
#         'itsdangerous==1.1.0',
#         'Jinja2==2.11.3',
#         'joblib==0.15.1',
#         'Keras-Applications==1.0.8',
#         'Keras-Preprocessing==1.1.0',
#         'kiwisolver==1.2.0',
#         'Markdown==3.2.1',
#         'MarkupSafe==1.1.1',
#         'matplotlib==3.3.2',
#         'mock==4.0.2',
#         'more-itertools==8.4.0',
#         'mpmath==1.1.0',
#         'numpy==1.18.2',
#         'NZMATH==1.1.0',
#         'oauthlib==3.1.0',
#         'opt-einsum==3.2.0',
#         'packaging==20.4',
#         'pandas==1.0.5',
#         'Pillow==8.1.0',
#         'pluggy==0.13.1',
#         'promise==2.3',
#         'protobuf==3.15.5',
#         'psutil==5.8.0',
#         'py==1.9.0',
#         'pyasn1==0.4.8',
#         'pyasn1-modules==0.2.8',
#         'pyparsing==2.4.7',
#         'pytest==5.4.3',
#         'python-dateutil==2.8.1',
#         'pytz==2020.1',
#         'requests==2.23.0',
#         'requests-oauthlib==1.3.0',
#         'rsa==4.0',
#         'scikit-learn==0.23.1',
#         'scipy==1.4.1',
#         'six==1.14.0',
#         'sympy==1.6',
#         'tensorboard==2.1.1',
#         'tensorflow==2.2.0rc0',
#         'tensorflow-datasets==4.2.0',
#         'tensorflow-estimator==2.1.0',
#         'tensorflow-metadata==0.28.0',
#         'termcolor==1.1.0',
#         'threadpoolctl==2.1.0',
#         'torch==1.8.1',
#         'torchtext==0.9.1',
#         'tqdm==4.59.0',
#         'typing-extensions==3.7.4.3',
#         'urllib3==1.25.9',
#         'wcwidth==0.2.5',
#         'Werkzeug==1.0.1',
#         'wrapt==1.12.1',
#         'xmlrunner==1.7.7',
#         'zipp==3.1.0'
#     ],



